"""Init for management of ISY Programs."""
import asyncio
from xml.dom import minidom

from dateutil import parser

from ..constants import (
    ATTR_ID,
    ATTR_PARENT,
    ATTR_STATUS,
    EMPTY_TIME,
    TAG_ENABLED,
    TAG_FOLDER,
    TAG_NAME,
    TAG_PRGM_FINISH,
    TAG_PRGM_RUN,
    TAG_PRGM_RUNNING,
    TAG_PRGM_STATUS,
    TAG_PROGRAM,
    UPDATE_INTERVAL,
    XML_OFF,
    XML_ON,
    XML_TRUE,
)
from ..exceptions import XML_ERRORS, XML_PARSE_ERROR
from ..helpers import attr_from_element, now, value_from_xml
from ..logging import _LOGGER
from ..nodes import NodeIterator as ProgramIterator
from .folder import Folder
from .program import Program


class Programs:
    """
    This class handles the ISY programs.

    This class can be used as a dictionary
    to navigate through the controller's structure to objects of type
    :class:`pyisy.programs.Program` and :class:`pyisy.programs.Folder`
    (when requested) that represent objects on the controller.

    |  isy: The ISY device class
    |  root: Program/Folder ID representing the current level of navigation.
    |  addresses: List of program and folder IDs.
    |  pnames: List of the program and folder names.
    |  pparents: List of the program and folder parent IDs.
    |  pobjs: List of program and folder objects.
    |  ptypes: List of the program and folder types.
    |  xml: XML string from the controller detailing the programs and folders.

    :ivar all_lower_programs: A list of all programs below the current
                            navigation level. Does not return folders.
    :ivar children: A list of the children immediately below the current
                    navigation level.
    :ivar leaf: The child object representing the current item in navigation.
                This is useful for getting a folder to act as a program.
    :ivar name: The name of the program at the current level of navigation.
    """

    def __init__(
        self,
        isy,
        root=None,
        addresses=None,
        pnames=None,
        pparents=None,
        pobjs=None,
        ptypes=None,
        xml=None,
    ):
        """Initialize the Programs ISY programs manager class."""
        self.isy = isy
        self.root = root

        self.addresses = []
        self.pnames = []
        self.pparents = []
        self.pobjs = []
        self.ptypes = []

        if xml is not None:
            self.parse(xml)
            return

        self.addresses = addresses
        self.pnames = pnames
        self.pparents = pparents
        self.pobjs = pobjs
        self.ptypes = ptypes

    def __str__(self):
        """Return a string representation of the program manager."""
        if self.root is None:
            return "Folder <root>"
        ind = self.addresses.index(self.root)
        if self.ptypes[ind] == TAG_FOLDER:
            return f"Folder ({self.root})"
        if self.ptypes[ind] == TAG_PROGRAM:
            return f"Program ({self.root})"
        return ""

    def __repr__(self):
        """Return a string showing the hierarchy of the program manager."""
        # get and sort children
        folders = []
        programs = []
        for child in self.children:
            if child[0] == TAG_FOLDER:
                folders.append(child)
            elif child[0] == TAG_PROGRAM:
                programs.append(child)

        # initialize data
        folders.sort(key=lambda x: x[1])
        programs.sort(key=lambda x: x[1])
        out = str(self) + "\n"

        # format folders
        for fold in folders:
            fold_obj = self[fold[2]]
            out += f"  + {fold[1]}: Folder({fold[2]})\n"
            for line in repr(fold_obj).split("\n")[1:]:
                out += f"  |   {line}\n"
            out += "  -\n"

        # format programs
        for prog in programs:
            out += f"  {prog[1]}: {self[prog[2]]}\n"

        return out

    def __iter__(self):
        """
        Return an iterator that iterates through all the programs.

        Does not iterate folders. Only Programs that are beneath the current
        folder in navigation.
        """
        iter_data = self.all_lower_programs
        return ProgramIterator(self, iter_data, delta=1)

    def __reversed__(self):
        """Return an iterator that goes in reverse order."""
        iter_data = self.all_lower_programs
        return ProgramIterator(self, iter_data, delta=-1)

    def update_received(self, xmldoc):
        """Update programs from EventStream message."""
        # pylint: disable=attribute-defined-outside-init
        xml = xmldoc.toxml()
        address = value_from_xml(xmldoc, ATTR_ID).zfill(4)
        try:
            pobj = self.get_by_id(address).leaf
        except ValueError:
            _LOGGER.warning(
                "ISY received program update for new program; reload the module to update"
            )
            return  # this is a new program that hasn't been registered

        if not isinstance(pobj, Program):
            return

        new_status = False

        if f"<{TAG_PRGM_STATUS}>" in xml:
            status = value_from_xml(xmldoc, TAG_PRGM_STATUS)
            if status == "21":
                pobj.ran_then += 1
                new_status = True
            elif status == "31":
                pobj.ran_else += 1

        if f"<{TAG_PRGM_RUN}>" in xml:
            pobj.last_run = parser.parse(value_from_xml(xmldoc, TAG_PRGM_RUN))

        if f"<{TAG_PRGM_FINISH}>" in xml:
            pobj.last_finished = parser.parse(value_from_xml(xmldoc, TAG_PRGM_FINISH))

        if XML_ON in xml or XML_OFF in xml:
            pobj.enabled = XML_ON in xml

        # Update Status last and make sure the change event fires, but only once.
        if pobj.status != new_status:
            pobj.status = new_status
        else:
            # Status didn't change, but something did, so fire the event.
            pobj.status_events.notify(new_status)

        _LOGGER.debug("ISY Updated Program: %s", address)

    def parse(self, xml):
        """
        Parse the XML from the controller and updates the state of the manager.

        xml: XML string from the controller.
        """
        try:
            xmldoc = minidom.parseString(xml)
        except XML_ERRORS:
            _LOGGER.error("%s: Programs, programs not loaded", XML_PARSE_ERROR)
            return

        plastup = now()

        # get nodes
        features = xmldoc.getElementsByTagName(TAG_PROGRAM)
        for feature in features:
            # id, name, and status
            address = attr_from_element(feature, ATTR_ID)
            pname = value_from_xml(feature, TAG_NAME)

            _LOGGER.debug("Parsing Program/Folder: %s [%s]", pname, address)

            pparent = attr_from_element(feature, ATTR_PARENT)
            pstatus = attr_from_element(feature, ATTR_STATUS) == XML_TRUE

            if attr_from_element(feature, TAG_FOLDER) == XML_TRUE:
                # folder specific parsing
                ptype = TAG_FOLDER
                data = {"pstatus": pstatus, "plastup": plastup}

            else:
                # program specific parsing
                ptype = TAG_PROGRAM

                # last run time
                plastrun = value_from_xml(feature, "lastRunTime", EMPTY_TIME)
                if plastrun != EMPTY_TIME:
                    plastrun = parser.parse(plastrun)

                # last finish time
                plastfin = value_from_xml(feature, "lastFinishTime", EMPTY_TIME)
                if plastfin != EMPTY_TIME:
                    plastfin = parser.parse(plastfin)

                # enabled, run at startup, running
                penabled = bool(attr_from_element(feature, TAG_ENABLED) == XML_TRUE)
                pstartrun = bool(attr_from_element(feature, "runAtStartup") == XML_TRUE)
                prunning = bool(attr_from_element(feature, TAG_PRGM_RUNNING) != "idle")

                # create data dictionary
                data = {
                    "pstatus": pstatus,
                    "plastrun": plastrun,
                    "plastfin": plastfin,
                    "penabled": penabled,
                    "pstartrun": pstartrun,
                    "prunning": prunning,
                    "plastup": plastup,
                }

            # add or update object if it already exists
            if address not in self.addresses:
                if ptype == TAG_FOLDER:
                    pobj = Folder(self, address, pname, **data)
                else:
                    pobj = Program(self, address, pname, **data)
                self.insert(address, pname, pparent, pobj, ptype)
            else:
                pobj = self.get_by_id(address).leaf
                asyncio.create_task(pobj.update(data=data))

        _LOGGER.info("ISY Loaded/Updated Programs")

    async def update(self, wait_time=UPDATE_INTERVAL, address=None):
        """
        Update the status of the programs and folders.

        |  wait_time: How long to wait before updating.
        |  address: The program ID to update.
        """
        await asyncio.sleep(wait_time)
        xml = await self.isy.conn.get_programs(address)

        if xml is not None:
            self.parse(xml)
        else:
            _LOGGER.warning("ISY Failed to update programs.")

    def insert(self, address, pname, pparent, pobj, ptype):
        """
        Insert a new program or folder into the manager.

        |  address: The ID of the program or folder.
        |  pname: The name of the program or folder.
        |  pparent: The parent of the program or folder.
        |  pobj: The object representing the program or folder.
        |  ptype: The type of the item being added (program/folder).
        """
        self.addresses.append(address)
        self.pnames.append(pname)
        self.pparents.append(pparent)
        self.ptypes.append(ptype)
        self.pobjs.append(pobj)

    def __getitem__(self, val):
        """
        Navigate through the hierarchy using names or IDs.

        |  val: Name or ID to navigate to.
        """
        try:
            self.addresses.index(val)
            fun = self.get_by_id
        except ValueError:
            try:
                self.pnames.index(val)
                fun = self.get_by_name
            except ValueError:
                try:
                    val = int(val)
                    fun = self.get_by_index
                except (TypeError, ValueError) as err:
                    raise KeyError("Unrecognized Key: " + str(val)) from err

        try:
            return fun(val)
        except (ValueError, KeyError, IndexError):
            return None

    def __setitem__(self, val, value):
        """Set the item value."""
        return None

    def get_by_name(self, val):
        """
        Get a child program/folder with the given name.

        |  val: The name of the child program/folder to look for.
        """
        for i in range(len(self.addresses)):
            if (self.root is None or self.pparents[i] == self.root) and self.pnames[
                i
            ] == val:
                return self.get_by_index(i)
        return None

    def get_by_id(self, address):
        """
        Get a program/folder with the given ID.

        |  address: The program/folder ID to look for.
        """
        i = self.addresses.index(address)
        return self.get_by_index(i)

    def get_by_index(self, i):
        """
        Get the program/folder at the given index.

        |  i: The program/folder index.
        """
        if self.ptypes[i] == TAG_FOLDER:
            return Programs(
                self.isy,
                self.addresses[i],
                self.addresses,
                self.pnames,
                self.pparents,
                self.pobjs,
                self.ptypes,
            )
        return self.pobjs[i]

    @property
    def children(self):
        """Return the children of the class."""
        out = []
        for ind, name in enumerate(self.pnames):
            if self.pparents[ind] == self.root:
                out.append((self.ptypes[ind], name, self.addresses[ind]))
        return out

    @property
    def leaf(self):
        """Return the leaf property."""
        if self.root is not None:
            ind = self.addresses.index(self.root)
            if self.pobjs[ind] is not None:
                return self.pobjs[ind]
        return self

    @property
    def name(self):
        """Return the name of the path."""
        if self.root is not None:
            ind = self.addresses.index(self.root)
            return self.pnames[ind]
        return ""

    @property
    def all_lower_programs(self):
        """Return all lower programs in a path."""
        output = []
        myname = self.name + "/"

        for dtype, name, ident in self.children:
            if dtype == TAG_PROGRAM:
                output.append((dtype, myname + name, ident))

            else:
                output += [
                    (dtype2, myname + name2, ident2)
                    for (dtype2, name2, ident2) in self[ident].all_lower_programs
                ]
        return output



"""ISY Program Folders."""
from ..constants import (
    ATTR_LAST_CHANGED,
    ATTR_LAST_UPDATE,
    ATTR_STATUS,
    CMD_DISABLE,
    CMD_ENABLE,
    CMD_RUN,
    CMD_RUN_ELSE,
    CMD_RUN_THEN,
    CMD_STOP,
    PROTO_FOLDER,
    TAG_ADDRESS,
    TAG_FOLDER,
    UPDATE_INTERVAL,
    URL_PROGRAMS,
)
from ..helpers import EventEmitter, now
from ..logging import _LOGGER


class Folder:
    """
    Object representing a program folder on the ISY device.

    |  programs: The folder manager object.
    |  address: The folder ID.
    |  pname: The folder name.
    |  pstatus: The current folder status.

    :ivar dtype: Returns the type of the object (folder).
    :ivar status: Watched property representing the current status of the
                  folder.
    """

    dtype = TAG_FOLDER

    def __init__(self, programs, address, pname, pstatus, plastup):
        """Initialize the Folder class."""
        self._id = address
        self._last_update = plastup
        self._last_changed = now()
        self._name = pname
        self._programs = programs
        self._status = pstatus
        self.isy = programs.isy
        self.status_events = EventEmitter()

    def __str__(self):
        """Return a string representation of the node."""
        return f"{type(self).__name__}({self._id})"

    @property
    def address(self):
        """Return the program or folder ID."""
        return self._id

    @property
    def last_changed(self):
        """Return the last time the program was changed in this module."""
        return self._last_changed

    @last_changed.setter
    def last_changed(self, value):
        """Set the last time the program was changed."""
        if self._last_changed != value:
            self._last_changed = value
        return self._last_changed

    @property
    def last_update(self):
        """Return the last time the program was updated."""
        return self._last_update

    @last_update.setter
    def last_update(self, value):
        """Set the last time the program was updated."""
        if self._last_update != value:
            self._last_update = value
        return self._last_update

    @property
    def leaf(self):
        """Get the leaf property."""
        return self

    @property
    def name(self):
        """Return the name of the Node."""
        return self._name

    @property
    def protocol(self):
        """Return the protocol for this entity."""
        return PROTO_FOLDER

    @property
    def status(self):
        """Return the current node state."""
        return self._status

    @status.setter
    def status(self, value):
        """Set the current node state and notify listeners."""
        if self._status != value:
            self._status = value
            self.status_events.notify(self._status)
        return self._status

    @property
    def status_feedback(self):
        """Return information for a status change event."""
        return {
            TAG_ADDRESS: self.address,
            ATTR_STATUS: self._status,
            ATTR_LAST_CHANGED: self._last_changed,
            ATTR_LAST_UPDATE: self._last_update,
        }

    async def update(self, wait_time=UPDATE_INTERVAL, data=None):
        """
        Update the status of the program.

        |  data: [optional] The data to update the folder with.
        |  wait_time: [optional] Seconds to wait before updating.
        """
        if data is not None:
            self._last_changed = now()
            self.status = data["pstatus"]
            return
        await self._programs.update(wait_time=wait_time, address=self._id)

    async def send_cmd(self, command):
        """Run the appropriate clause of the object."""
        req_url = self.isy.conn.compile_url([URL_PROGRAMS, str(self._id), command])
        result = await self.isy.conn.request(req_url)
        if not result:
            _LOGGER.warning('ISY could not call "%s" on program: %s', command, self._id)
            return False
        _LOGGER.debug('ISY ran "%s" on program: %s', command, self._id)
        if not self.isy.auto_update:
            await self.update()
        return True

    async def enable(self):
        """Send command to the program/folder to enable it."""
        return await self.send_cmd(CMD_ENABLE)

    async def disable(self):
        """Send command to the program/folder to enable it."""
        return await self.send_cmd(CMD_DISABLE)

    async def run(self):
        """Send a run command to the program/folder."""
        return await self.send_cmd(CMD_RUN)

    async def run_then(self):
        """Send a runThen command to the program/folder."""
        return await self.send_cmd(CMD_RUN_THEN)

    async def run_else(self):
        """Send a runElse command to the program/folder."""
        return await self.send_cmd(CMD_RUN_ELSE)

    async def stop(self):
        """Send a stop command to the program/folder."""
        return await self.send_cmd(CMD_STOP)



"""Manage variables from the ISY."""
from ..constants import (
    ATTR_INIT,
    ATTR_LAST_CHANGED,
    ATTR_LAST_UPDATE,
    ATTR_PRECISION,
    ATTR_SET,
    ATTR_STATUS,
    ATTR_TS,
    PROTO_INT_VAR,
    PROTO_STATE_VAR,
    TAG_ADDRESS,
    URL_VARIABLES,
    VAR_INTEGER,
)
from ..helpers import EventEmitter, now
from ..logging import _LOGGER


class Variable:
    """
    Object representing a variable on the controller.

    |  variables: The variable manager object.
    |  vid: List of variable IDs.
    |  vtype: List of variable types.
    |  init: List of values that variables initialize to when the controller
             starts.
    |  val: The current variable value.
    |  ts: The timestamp for the last time the variable was edited.

    :ivar init: Watched property that represents the value the variable
                initializes to when the controller boots.
    :ivar lastEdit: Watched property that indicates the last time the variable
                    was edited.
    :ivar val: Watched property that represents the value of the variable.
    """

    def __init__(self, variables, vid, vtype, vname, init, status, timestamp, prec):
        """Initialize a Variable class."""
        super().__init__()
        self._id = vid
        self._init = init
        self._last_edited = timestamp
        self._last_update = now()
        self._last_changed = now()
        self._name = vname
        self._prec = prec
        self._status = status
        self._type = vtype
        self._variables = variables
        self.isy = variables.isy
        self.status_events = EventEmitter()

    def __str__(self):
        """Return a string representation of the variable."""
        return f"Variable(type={self._type}, id={self._id}, value={self.status}, init={self.init})"

    def __repr__(self):
        """Return a string representation of the variable."""
        return str(self)

    @property
    def address(self):
        """Return the formatted Variable Type and ID."""
        return f"{self._type}.{self._id}"

    @property
    def init(self):
        """Return the initial state."""
        return self._init

    @init.setter
    def init(self, value):
        """Set the initial state and notify listeners."""
        if self._init != value:
            self._init = value
            self._last_changed = now()
            self.status_events.notify(self.status_feedback)
        return self._init

    @property
    def last_changed(self):
        """Return the UTC Time of the last status change for this node."""
        return self._last_changed

    @property
    def last_edited(self):
        """Return the last edit time."""
        return self._last_edited

    @last_edited.setter
    def last_edited(self, value):
        """Set the last edited time."""
        if self._last_edited != value:
            self._last_edited = value
        return self._last_edited

    @property
    def last_update(self):
        """Return the UTC Time of the last update for this node."""
        return self._last_update

    @last_update.setter
    def last_update(self, value):
        """Set the last update time."""
        if self._last_update != value:
            self._last_update = value
        return self._last_update

    @property
    def protocol(self):
        """Return the protocol for this entity."""
        return PROTO_INT_VAR if self._type == VAR_INTEGER else PROTO_STATE_VAR

    @property
    def name(self):
        """Return the Variable Name."""
        return self._name

    @property
    def prec(self):
        """Return the Variable Precision."""
        return self._prec

    @prec.setter
    def prec(self, value):
        """Set the current node state and notify listeners."""
        if self._prec != value:
            self._prec = value
            self._last_changed = now()
            self.status_events.notify(self.status_feedback)
        return self._prec

    @property
    def status(self):
        """Return the current node state."""
        return self._status

    @status.setter
    def status(self, value):
        """Set the current node state and notify listeners."""
        if self._status != value:
            self._status = value
            self._last_changed = now()
            self.status_events.notify(self.status_feedback)
        return self._status

    @property
    def status_feedback(self):
        """Return information for a status change event."""
        return {
            TAG_ADDRESS: self.address,
            ATTR_STATUS: self._status,
            ATTR_INIT: self._init,
            ATTR_PRECISION: self._prec,
            ATTR_TS: self._last_edited,
            ATTR_LAST_CHANGED: self._last_changed,
            ATTR_LAST_UPDATE: self._last_update,
        }

    @property
    def vid(self):
        """Return the Variable ID."""
        return self._id

    async def update(self, wait_time=0):
        """
        Update the object with the variable's parameters from the controller.

        |  wait_time: Seconds to wait before updating.
        """
        self._last_update = now()
        await self._variables.update(wait_time)

    async def set_init(self, val):
        """
        Set the initial value for the variable after the controller boots.

        |  val: The value to have the variable initialize to.
        """
        if val is None:
            raise ValueError("Variable init must be an integer. Got None.")
        self.set_value(val, True)

    async def set_value(self, val, init=False):
        """
        Set the value of the variable.

        |  val: The value to set the variable to.
        """
        if val is None:
            raise ValueError("Variable value must be an integer. Got None.")
        req_url = self.isy.conn.compile_url(
            [
                URL_VARIABLES,
                ATTR_INIT if init else ATTR_SET,
                str(self._type),
                str(self._id),
                str(val),
            ]
        )
        if not await self.isy.conn.request(req_url):
            _LOGGER.warning(
                "ISY could not set variable%s: %s.%s",
                " init value" if init else "",
                str(self._type),
                str(self._id),
            )
            return
        _LOGGER.debug(
            "ISY set variable%s: %s.%s",
            " init value" if init else "",
            str(self._type),
            str(self._id),
        )
        if not self.isy.auto_update:
            await self.update()



"""Representation of groups (scenes) from an ISY."""
from ..constants import (
    FAMILY_GENERIC,
    INSTEON_STATELESS_NODEDEFID,
    ISY_VALUE_UNKNOWN,
    PROTO_GROUP,
)
from ..helpers import now
from .nodebase import NodeBase


class Group(NodeBase):
    """
    Interact with ISY groups (scenes).

    |  nodes: The node manager object.
    |  address: The node ID.
    |  name: The node name.
    |  members: List of the members in this group.
    |  controllers: List of the controllers in this group.
    |  spoken: The string of the Notes Spoken field.

    :ivar has_children: Boolean value indicating that group has no children.
    :ivar members: List of the members of this group.
    :ivar controllers: List of the controllers of this group.
    :ivar name: The name of this group.
    :ivar status: Watched property indicating the status of the group.
    :ivar group_all_on: Watched property indicating if all devices in group are on.
    """

    def __init__(
        self,
        nodes,
        address,
        name,
        members=None,
        controllers=None,
        family_id=FAMILY_GENERIC,
        pnode=None,
        flag=0,
    ):
        """Initialize a Group class."""
        self._all_on = False
        self._controllers = controllers or []
        self._members = members or []
        super().__init__(
            nodes, address, name, 0, family_id=family_id, pnode=pnode, flag=flag
        )

        # listen for changes in children
        self._members_handlers = [
            self._nodes[m].status_events.subscribe(self.update_callback)
            for m in self.members
        ]

        # get and update the status
        self._update()

    def __del__(self):
        """Cleanup event handlers before deleting."""
        for handler in self._members_handlers:
            handler.unsubscribe()

    @property
    def controllers(self):
        """Get the controller nodes of the scene/group."""
        return self._controllers

    @property
    def group_all_on(self):
        """Return the current node state."""
        return self._all_on

    @group_all_on.setter
    def group_all_on(self, value):
        """Set the current node state and notify listeners."""
        if self._all_on != value:
            self._all_on = value
            self._last_changed = now()
            # Re-publish the current status. Let users pick up the all on change.
            self.status_events.notify(self._status)
        return self._all_on

    @property
    def members(self):
        """Get the members of the scene/group."""
        return self._members

    @property
    def protocol(self):
        """Return the protocol for this entity."""
        return PROTO_GROUP

    async def update(self, event=None, wait_time=0, xmldoc=None):
        """Update the group with values from the controller."""
        return self._update(event, wait_time, xmldoc)

    def _update(self, event=None, wait_time=0, xmldoc=None):
        """Update the group with values from the controller."""
        self._last_update = now()
        valid_nodes = [
            node
            for node in self.members
            if (
                self._nodes[node].status is not None
                and self._nodes[node].status != ISY_VALUE_UNKNOWN
                and self._nodes[node].node_def_id not in INSTEON_STATELESS_NODEDEFID
            )
        ]
        on_nodes = [node for node in valid_nodes if int(self._nodes[node].status) > 0]
        if on_nodes:
            self.group_all_on = len(on_nodes) == len(valid_nodes)
            self.status = 255
            return
        self.status = 0
        self.group_all_on = False

    def update_callback(self, event=None):
        """Handle synchronous callbacks for subscriber events."""
        self._update(event)



"""Connection to the ISY."""
import asyncio
import ssl
import sys
from urllib.parse import quote, urlencode

import aiohttp

from .constants import (
    ATTR_ACTION,
    METHOD_GET,
    SYSTEM_BUSY,
    SYSTEM_IDLE,
    SYSTEM_NOT_BUSY,
    SYSTEM_STATUS,
    URL_CLOCK,
    URL_CONFIG,
    URL_DEFINITIONS,
    URL_MEMBERS,
    URL_NETWORK,
    URL_NODES,
    URL_PING,
    URL_PROGRAMS,
    URL_RESOURCES,
    URL_STATUS,
    URL_SUBFOLDERS,
    URL_VARIABLES,
    VAR_INTEGER,
    VAR_STATE,
    XML_FALSE,
    XML_TRUE,
)
from .exceptions import ISYConnectionError, ISYInvalidAuthError
from .helpers import EventEmitter, value_from_xml
from .logging import _LOGGER, enable_logging

MAX_HTTPS_CONNECTIONS_ISY = 2
MAX_HTTP_CONNECTIONS_ISY = 5
MAX_HTTPS_CONNECTIONS_IOX = 20
MAX_HTTP_CONNECTIONS_IOX = 50

MAX_RETRIES = 5
RETRY_BACKOFF = [0.01, 0.10, 0.25, 1, 2]  # Seconds

HTTP_OK = 200  # Valid request received, will run it
HTTP_UNAUTHORIZED = 401  # User authentication failed
HTTP_NOT_FOUND = 404  # Unrecognized request received and ignored
HTTP_SERVICE_UNAVAILABLE = 503  # Valid request received, system too busy to run it

HTTP_TIMEOUT = 30

HTTP_HEADERS = {
    "Connection": "keep-alive",
    "Keep-Alive": "5000",
    "Accept-Encoding": "gzip, deflate",
}

EMPTY_XML_RESPONSE = '<?xml version="1.0" encoding="UTF-8"?>'


class ISYSystemStatus:
    """Event manager class for ISY System Status."""

    _event = asyncio.Event()
    _status = SYSTEM_IDLE

    def __init__(self):
        """Initialize a system status class."""
        self.status_events = EventEmitter()
        self._event.set()

    @property
    def status(self):
        """Return the system status property."""
        return self._status

    @status.setter
    def status(self, val):
        """Update the system status property."""
        self._status = val
        if val == SYSTEM_BUSY:
            self._event.clear()
        elif val in (SYSTEM_IDLE, SYSTEM_NOT_BUSY):
            self._event.set()
        else:
            raise ISYConnectionError("ISY status unknown")

    async def is_ready(self):
        """Wait for the ISY to be ready."""
        return await self._event.wait()

    def change_received(self, xmldoc):
        """Handle System Status events from an event stream message."""
        action = value_from_xml(xmldoc, ATTR_ACTION)
        if not action or action not in SYSTEM_STATUS:
            return
        self.status = action
        self.status_events.notify(action)


class Connection:
    """Connection object to manage connection to and interaction with ISY."""

    def __init__(
        self,
        address,
        port,
        username,
        password,
        use_https=False,
        tls_ver=1.1,
        webroot="",
        websession=None,
    ):
        """Initialize the Connection object."""
        if len(_LOGGER.handlers) == 0:
            enable_logging(add_null_handler=True)

        self._address = address
        self._port = port
        self._username = username
        self._password = password
        self._auth = aiohttp.BasicAuth(self._username, self._password)
        self._webroot = webroot.rstrip("/")
        self.req_session = websession
        self._tls_ver = tls_ver
        self.use_https = use_https
        self._url = f"http{'s' if self.use_https else ''}://{self._address}:{self._port}{self._webroot}"
        self.system_status = ISYSystemStatus()

        self.semaphore = asyncio.Semaphore(
            MAX_HTTPS_CONNECTIONS_ISY if use_https else MAX_HTTP_CONNECTIONS_ISY
        )

        if websession is None:
            websession = get_new_client_session(use_https, tls_ver)
        self.req_session = websession
        self.sslcontext = get_sslcontext(use_https, tls_ver)

    async def test_connection(self):
        """Test the connection and get the config for the ISY."""
        config = await self.get_config(retries=None)
        if not config:
            _LOGGER.error("Could not connect to the ISY with the parameters provided.")
            raise ISYConnectionError()
        return config

    def increase_available_connections(self):
        """Increase the number of allowed connections for newer hardware."""
        _LOGGER.debug("Increasing available simultaneous connections")
        self.semaphore = asyncio.Semaphore(
            MAX_HTTPS_CONNECTIONS_IOX if self.use_https else MAX_HTTP_CONNECTIONS_IOX
        )

    async def close(self):
        """Cleanup connections and prepare for exit."""
        await self.req_session.close()

    @property
    def connection_info(self):
        """Return the connection info required to connect to the ISY."""
        connection_info = {}
        connection_info["auth"] = self._auth.encode()
        connection_info["addr"] = self._address
        connection_info["port"] = int(self._port)
        connection_info["passwd"] = self._password
        connection_info["webroot"] = self._webroot
        if self.use_https and self._tls_ver:
            connection_info["tls"] = self._tls_ver

        return connection_info

    @property
    def url(self):
        """Return the full connection url."""
        return self._url

    # COMMON UTILITIES
    def compile_url(self, path, query=None):
        """Compile the URL to fetch from the ISY."""
        url = self.url
        if path is not None:
            url += "/rest/" + "/".join([quote(item) for item in path])

        if query is not None:
            url += "?" + urlencode(query)

        return url

    async def request(self, url, retries=0, ok404=False, delay=0):
        """Execute request to ISY REST interface."""
        _LOGGER.debug("ISY Request: %s", url)
        if delay:
            await asyncio.sleep(delay)
        try:
            async with self.semaphore, self.req_session.get(
                url,
                auth=self._auth,
                headers=HTTP_HEADERS,
                timeout=HTTP_TIMEOUT,
                ssl=self.sslcontext,
            ) as res:
                endpoint = url.split("rest", 1)[1]
                if res.status == HTTP_OK:
                    _LOGGER.debug("ISY Response Received: %s", endpoint)
                    results = await res.text(encoding="utf-8", errors="ignore")
                    if results != EMPTY_XML_RESPONSE:
                        return results
                    _LOGGER.debug("Invalid empty XML returned: %s", endpoint)
                    res.release()
                if res.status == HTTP_NOT_FOUND:
                    if ok404:
                        _LOGGER.debug("ISY Response Received %s", endpoint)
                        res.release()
                        return ""
                    _LOGGER.error(
                        "ISY Reported an Invalid Command Received %s", endpoint
                    )
                    res.release()
                    # ISY may report 404 error for valid command if busy
                    if self.system_status.status != SYSTEM_BUSY:
                        return None
                if res.status == HTTP_UNAUTHORIZED:
                    _LOGGER.error("Invalid credentials provided for ISY connection.")
                    res.release()
                    raise ISYInvalidAuthError(
                        "Invalid credentials provided for ISY connection."
                    )
                if res.status == HTTP_SERVICE_UNAVAILABLE:
                    _LOGGER.warning("ISY too busy to process request %s", endpoint)
                    res.release()

        except asyncio.TimeoutError:
            _LOGGER.warning("Timeout while trying to connect to the ISY.")
        except (
            aiohttp.ClientOSError,
            aiohttp.ServerDisconnectedError,
        ):
            _LOGGER.debug("ISY not ready or closed connection.")
        except aiohttp.ClientResponseError as err:
            _LOGGER.error(
                "Client Response Error from ISY: %s %s.", err.status, err.message
            )
        except aiohttp.ClientError as err:
            _LOGGER.error(
                "ISY Could not receive response from device because of a network issue: %s",
                type(err),
            )

        if retries is None:
            raise ISYConnectionError()
        if retries < MAX_RETRIES:
            if self.system_status.status == SYSTEM_BUSY:
                _LOGGER.debug("ISY is busy, waiting for system to be ready")
                try:
                    await asyncio.wait_for(self.system_status.is_ready(), timeout=5.0)
                except asyncio.TimeoutError as exc:
                    raise ISYConnectionError() from exc
            else:
                _LOGGER.debug(
                    "Retrying ISY Request in %ss, retry %s.",
                    RETRY_BACKOFF[retries],
                    retries + 1,
                )
                # sleep to allow the ISY to catch up
                await asyncio.sleep(RETRY_BACKOFF[retries])
            # recurse to try again
            retry_result = await self.request(url, retries + 1, ok404=ok404)
            return retry_result
        # fail for good
        _LOGGER.error(
            "Bad ISY Request: (%s) Failed after %s retries.",
            url,
            retries,
        )
        return None

    async def ping(self):
        """Test connection to the ISY and return True if alive."""
        req_url = self.compile_url([URL_PING])
        result = await self.request(req_url, ok404=True)
        return result is not None

    async def get_description(self):
        """Fetch the services description from the ISY."""
        url = "https://" if self.use_https else "http://"
        url += f"{self._address}:{self._port}{self._webroot}/desc"
        result = await self.request(url)
        return result

    async def get_config(self, retries=0):
        """Fetch the configuration from the ISY."""
        req_url = self.compile_url([URL_CONFIG])
        result = await self.request(req_url, retries=retries)
        return result

    async def get_programs(self, address=None):
        """Fetch the list of programs from the ISY."""
        addr = [URL_PROGRAMS]
        if address is not None:
            addr.append(str(address))
        req_url = self.compile_url(addr, {URL_SUBFOLDERS: XML_TRUE})
        result = await self.request(req_url)
        return result

    async def get_nodes(self):
        """Fetch the list of nodes/groups/scenes from the ISY."""
        req_url = self.compile_url([URL_NODES], {URL_MEMBERS: XML_FALSE})
        result = await self.request(req_url)
        return result

    async def get_status(self):
        """Fetch the status of nodes/groups/scenes from the ISY."""
        req_url = self.compile_url([URL_STATUS])
        result = await self.request(req_url)
        return result

    async def get_variable_defs(self):
        """Fetch the list of variables from the ISY."""
        req_list = [
            [URL_VARIABLES, URL_DEFINITIONS, VAR_INTEGER],
            [URL_VARIABLES, URL_DEFINITIONS, VAR_STATE],
        ]
        req_urls = [self.compile_url(req) for req in req_list]
        results = await asyncio.gather(
            *[self.request(req_url) for req_url in req_urls], return_exceptions=True
        )
        return results

    async def get_variables(self):
        """Fetch the variable details from the ISY to update local copy."""
        req_list = [
            [URL_VARIABLES, METHOD_GET, VAR_INTEGER],
            [URL_VARIABLES, METHOD_GET, VAR_STATE],
        ]
        req_urls = [self.compile_url(req) for req in req_list]
        results = await asyncio.gather(
            *[self.request(req_url) for req_url in req_urls], return_exceptions=True
        )
        results = [r for r in results if r is not None]  # Strip any bad requests.
        result = "".join(results)
        result = result.replace(
            '</vars><?xml version="1.0" encoding="UTF-8"?><vars>', ""
        )
        return result

    async def get_network(self):
        """Fetch the list of network resources from the ISY."""
        req_url = self.compile_url([URL_NETWORK, URL_RESOURCES])
        result = await self.request(req_url)
        return result

    async def get_time(self):
        """Fetch the system time info from the ISY."""
        req_url = self.compile_url([URL_CLOCK])
        result = await self.request(req_url)
        return result


def get_new_client_session(use_https, tls_ver=1.1):
    """Create a new Client Session for Connecting."""
    if use_https:
        if not can_https(tls_ver):
            raise (
                ValueError(
                    "PyISY could not connect to the ISY. "
                    "Check log for SSL/TLS error."
                )
            )

        return aiohttp.ClientSession(cookie_jar=aiohttp.CookieJar(unsafe=True))

    return aiohttp.ClientSession()


def get_sslcontext(use_https, tls_ver=1.1):
    """Create an SSLContext object to use for the connections."""
    if not use_https:
        return None
    if tls_ver == 1.1:
        context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_1)
    elif tls_ver == 1.2:
        context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)

    # Allow older ciphers for older ISYs
    context.set_ciphers(
        "DEFAULT:!aNULL:!eNULL:!MD5:!3DES:!DES:!RC4:!IDEA:!SEED:!aDSS:!SRP:!PSK"
    )
    return context


def can_https(tls_ver):
    """
    Verify minimum requirements to use an HTTPS connection.

    Returns boolean indicating whether HTTPS is available.
    """
    output = True

    # check python version
    if sys.version_info < (3, 7):
        _LOGGER.error("PyISY cannot use HTTPS: Invalid Python version. See docs.")
        output = False

    # check that Python was compiled against correct OpenSSL lib
    if "PROTOCOL_TLSv1_1" not in dir(ssl):
        _LOGGER.error(
            "PyISY cannot use HTTPS: Compiled against old OpenSSL library. See docs."
        )
        output = False

    # check the requested TLS version
    if tls_ver not in [1.1, 1.2]:
        _LOGGER.error(
            "PyISY cannot use HTTPS: Only TLS 1.1 and 1.2 are supported by the ISY controller."
        )
        output = False

    return output



"""Strings for Event Stream Requests."""

# Subscribe Message
SUB_MSG = {
    "head": """POST /services HTTP/1.1
Host: {addr}:{port}{webroot}
Authorization: {auth}
Content-Length: {length}
Content-Type: text/xml; charset="utf-8"
SOAPAction: urn:udi-com:device:X_Insteon_Lighting_Service:1#Subscribe\r
\r
""",
    "body": """<s:Envelope><s:Body>
<u:Subscribe xmlns:u="urn:udi-com:service:X_Insteon_Lighting_Service:1">
<reportURL>REUSE_SOCKET</reportURL>
<duration>infinite</duration>
</u:Subscribe></s:Body></s:Envelope>
\r
""",
}

# Unsubscribe Message
UNSUB_MSG = {
    "head": """POST /services HTTP/1.1
Host: {addr}:{port}{webroot}
Authorization: {auth}
Content-Length: {length}
Content-Type: text/xml; charset="utf-8"
SOAPAction: urn:udi-com:device:X_Insteon_Lighting_Service:1#Unsubscribe\r
\r
""",
    "body": """<s:Envelope><s:Body>
<u:Unsubscribe xmlns:u="urn:udi-com:service:X_Insteon_Lighting_Service:1">
<SID>{sid}</SID>
</u:Unsubscribe></s:Body></s:Envelope>
\r
""",
}

# Resubscribe Message
RESUB_MSG = {
    "head": """POST /services HTTP/1.1
Host: {addr}:{port}{webroot}
Authorization: {auth}
Content-Length: {length}
Content-Type: text/xml; charset="utf-8"
SOAPAction: urn:udi-com:device:X_Insteon_Lighting_Service:1#Subscribe\r
\r
""",
    "body": """<s:Envelope><s:Body>
<u:Subscribe xmlns:u="urn:udi-com:service:X_Insteon_Lighting_Service:1">
<reportURL>REUSE_SOCKET</reportURL>
<duration>infinite</duration>
<SID>{sid}</SID>
</u:Subscribe></s:Body></s:Envelope>
\r
""",
}



"""Representation of a program from the ISY."""
from ..constants import (
    CMD_DISABLE_RUN_AT_STARTUP,
    CMD_ENABLE_RUN_AT_STARTUP,
    PROTO_PROGRAM,
    TAG_PROGRAM,
    UPDATE_INTERVAL,
)
from .folder import Folder


class Program(Folder):
    """
    Class representing a program on the ISY controller.

    |  programs: The program manager object.
    |  address: The ID of the program.
    |  pname: The name of the program.
    |  pstatus: The current status of the program.
    |  plastup: The last time the program was updated.
    |  plastrun: The last time the program was run.
    |  plastfin: The last time the program finished running.
    |  penabled: Boolean value showing if the program is enabled on the
                 controller.
    |  pstartrun: Boolean value showing if the if the program runs on
                  controller start up.
    |  prunning: Boolean value showing if the current program is running
                 on the controller.
    """

    dtype = TAG_PROGRAM

    def __init__(
        self,
        programs,
        address,
        pname,
        pstatus,
        plastup,
        plastrun,
        plastfin,
        penabled,
        pstartrun,
        prunning,
    ):
        """Initialize a Program class."""
        super().__init__(programs, address, pname, pstatus, plastup)
        self._enabled = penabled
        self._last_finished = plastfin
        self._last_run = plastrun
        self._ran_else = 0
        self._ran_then = 0
        self._run_at_startup = pstartrun
        self._running = prunning

    @property
    def enabled(self):
        """Return if the program is enabled on the controller."""
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        """Set if the program is enabled on the controller."""
        if self._enabled != value:
            self._enabled = value
        return self._enabled

    @property
    def last_finished(self):
        """Return the last time the program finished running."""
        return self._last_finished

    @last_finished.setter
    def last_finished(self, value):
        """Set the last time the program finished running."""
        if self._last_finished != value:
            self._last_finished = value
        return self._last_finished

    @property
    def last_run(self):
        """Return the last time the program was run."""
        return self._last_run

    @last_run.setter
    def last_run(self, value):
        """Set the last time the program was run."""
        if self._last_run != value:
            self._last_run = value
        return self._last_run

    @property
    def protocol(self):
        """Return the protocol for this entity."""
        return PROTO_PROGRAM

    @property
    def ran_else(self):
        """Return the Ran Else property for this program."""
        return self._ran_else

    @ran_else.setter
    def ran_else(self, value):
        """Set the Ran Else property for this program."""
        if self._ran_else != value:
            self._ran_else = value
        return self._ran_else

    @property
    def ran_then(self):
        """Return the Ran Then property for this program."""
        return self._ran_then

    @ran_then.setter
    def ran_then(self, value):
        """Set the Ran Then property for this program."""
        if self._ran_then != value:
            self._ran_then = value
        return self._ran_then

    @property
    def run_at_startup(self):
        """Return if the program runs on controller start up."""
        return self._run_at_startup

    @run_at_startup.setter
    def run_at_startup(self, value):
        """Set if the program runs on controller start up."""
        if self._run_at_startup != value:
            self._run_at_startup = value
        return self._run_at_startup

    @property
    def running(self):
        """Return if the current program is running on the controller."""
        return self._running

    @running.setter
    def running(self, value):
        """Set if the current program is running on the controller."""
        if self._running != value:
            self._running = value
        return self._running

    async def update(self, wait_time=UPDATE_INTERVAL, data=None):
        """
        Update the program with values on the controller.

        |  wait_time: [optional] Seconds to wait before updating.
        |  data: [optional] Data to update the object with.
        """
        if data is not None:
            self._enabled = data["penabled"]
            self._last_finished = data["plastfin"]
            self._last_run = data["plastrun"]
            self._last_update = data["plastup"]
            self._run_at_startup = data["pstartrun"]
            self._running = (data["plastrun"] >= data["plastup"]) or data["prunning"]
            # Update Status last and make sure the change event fires, but only once.
            if self.status != data["pstatus"]:
                self.status = data["pstatus"]
            else:
                # Status didn't change, but something did, so fire the event.
                self.status_events.notify(self.status)
            return
        await self._programs.update(wait_time, address=self._id)

    async def enable_run_at_startup(self):
        """Send command to the program to enable it to run at startup."""
        return await self.send_cmd(CMD_ENABLE_RUN_AT_STARTUP)

    async def disable_run_at_startup(self):
        """Send command to the program to enable it to run at startup."""
        return await self.send_cmd(CMD_DISABLE_RUN_AT_STARTUP)



"""Exceptions used by the PyISY module."""
from xml.parsers.expat import ExpatError

XML_ERRORS = (AttributeError, KeyError, ValueError, TypeError, IndexError, ExpatError)
XML_PARSE_ERROR = "ISY Could not parse response, poorly formatted XML."


class ISYInvalidAuthError(Exception):
    """Invalid authorization credentials provided."""


class ISYConnectionError(Exception):
    """Invalid connection parameters provided."""


class ISYResponseParseError(Exception):
    """Error parsing a response provided by the ISY."""


class ISYStreamDataError(Exception):
    """Invalid data in the isy event stream."""


class ISYStreamDisconnected(ISYStreamDataError):
    """The isy has disconnected."""


class ISYMaxConnections(ISYStreamDisconnected):
    """The isy has disconnected because it reached maximum connections."""



"""Representation of ISY Nodes."""
from asyncio import sleep
from dataclasses import dataclass
import re
from xml.dom import minidom

from ..constants import (
    ATTR_ACTION,
    ATTR_CONTROL,
    ATTR_FLAG,
    ATTR_ID,
    ATTR_INSTANCE,
    ATTR_NODE_DEF_ID,
    ATTR_PRECISION,
    ATTR_UNIT_OF_MEASURE,
    DEFAULT_PRECISION,
    DEFAULT_UNIT_OF_MEASURE,
    DEV_MEMORY,
    DEV_WRITING,
    EVENT_PROPS_IGNORED,
    FAMILY_BRULTECH,
    FAMILY_NODESERVER,
    FAMILY_RCS,
    FAMILY_ZMATTER_ZWAVE,
    FAMILY_ZWAVE,
    INSTEON_RAMP_RATES,
    ISY_VALUE_UNKNOWN,
    NC_NODE_ENABLED,
    NC_NODE_ERROR,
    NODE_CHANGED_ACTIONS,
    NODE_IS_CONTROLLER,
    NODE_IS_ROOT,
    PROP_BATTERY_LEVEL,
    PROP_COMMS_ERROR,
    PROP_RAMP_RATE,
    PROP_STATUS,
    PROTO_INSTEON,
    PROTO_NODE_SERVER,
    PROTO_ZIGBEE,
    PROTO_ZWAVE,
    TAG_ADDRESS,
    TAG_DEVICE_TYPE,
    TAG_ENABLED,
    TAG_EVENT_INFO,
    TAG_FAMILY,
    TAG_FOLDER,
    TAG_FORMATTED,
    TAG_GROUP,
    TAG_LINK,
    TAG_NAME,
    TAG_NODE,
    TAG_PARENT,
    TAG_PRIMARY_NODE,
    TAG_TYPE,
    UOM_SECONDS,
    XML_TRUE,
)
from ..exceptions import XML_ERRORS, XML_PARSE_ERROR, ISYResponseParseError
from ..helpers import (
    EventEmitter,
    NodeProperty,
    ZWaveProperties,
    attr_from_element,
    attr_from_xml,
    parse_xml_properties,
    value_from_xml,
)
from ..logging import _LOGGER
from ..node_servers import NodeServers
from .group import Group
from .node import Node

MEMORY_REGEX = (
    r".*dbAddr=(?P<dbAddr>[A-F0-9x]*) \[(?P<value>[A-F0-9]{2})\] "
    r"cmd1=(?P<cmd1>[A-F0-9x]{4}) cmd2=(?P<cmd2>[A-F0-9x]{4})"
)


class Nodes:
    """
    This class handles the ISY nodes.

    This class can be used as a dictionary to
    navigate through the controller's structure to objects of type
    :class:`pyisy.nodes.Node` and :class:`pyisy.nodes.Group` that represent
    objects on the controller.

    |  isy: ISY class
    |  root: [optional] String representing the current navigation level's ID
    |  addresses: [optional] list of node ids
    |  nnames: [optional] list of node names
    |  nparents: [optional] list of node parents
    |  nobjs: [optional] list of node objects
    |  ntypes: [optional] list of node types
    |  xml: [optional] String of xml data containing the configuration data

    :ivar all_lower_nodes: Return all nodes beneath current level
    :ivar children: A list of the object's children.
    :ivar has_children: Indicates if object has children
    :ivar name: The name of the current folder in navigation.
    """

    def __init__(
        self,
        isy,
        root=None,
        addresses=None,
        nnames=None,
        nparents=None,
        nobjs=None,
        ntypes=None,
        xml=None,
    ):
        """Initialize the Nodes ISY Node Manager class."""
        self.isy = isy
        self.root = root

        self.addresses = []
        self.nnames = []
        self.nparents = []
        self.nobjs = []
        self.ntypes = []

        self.status_events = EventEmitter()

        if xml is not None:
            self.parse(xml)
            return

        self.addresses = addresses
        self.nnames = nnames
        self.nparents = nparents
        self.nobjs = nobjs
        self.ntypes = ntypes

    def __str__(self):
        """Return string representation of the nodes/folders/groups."""
        if self.root is None:
            return "Folder <root>"
        ind = self.addresses.index(self.root)
        if self.ntypes[ind] == TAG_FOLDER:
            return f"Folder ({self.root})"
        if self.ntypes[ind] == TAG_GROUP:
            return f"Group ({self.root})"
        return f"Node ({self.root})"

    def __repr__(self):
        """Create a pretty representation of the nodes/folders/groups."""
        # get and sort children
        folders = []
        groups = []
        nodes = []
        for child in self.children:
            if child[0] == TAG_FOLDER:
                folders.append(child)
            elif child[0] == TAG_GROUP:
                groups.append(child)
            elif child[0] == TAG_NODE:
                nodes.append(child)

        # initialize data
        folders.sort(key=lambda x: x[1])
        groups.sort(key=lambda x: x[1])
        nodes.sort(key=lambda x: x[1])
        out = (
            f"{self}\n"
            f"{self.__repr_folders__(folders)}"
            f"{self.__repr_groups__(groups)}"
            f"{self.__repr_nodes__(nodes)}"
        )
        return out

    def __repr_folders__(self, folders):
        """Return a representation of the folder structure."""
        out = ""
        for fold in folders:
            fold_obj = self[fold[2]]
            out += f"  + {fold[1]}: Folder({fold[2]})\n"
            for line in repr(fold_obj).split("\n")[1:]:
                out += f"  |   {line}\n"
            out += "  -\n"
        return out

    def __repr_groups__(self, groups):
        """Return a representation of the groups structure."""
        out = ""
        for group in groups:
            out += f"  + {group[1]}: Group({group[2]})\n"
            for member in self[group[2]].members:
                out += f"  |  {self[member].name}: Node({member})\n"
            out += "  |\n  -\n"
        return out

    def __repr_nodes__(self, nodes):
        """Return a representation of the nodes structure."""
        out = ""
        for node in nodes:
            has_children = node[2] in self.nparents
            out += f"  {'+ ' if has_children else ''}{node[1]}: Node({node[2]})\n"
            if has_children:
                for child in self.get_children(node[2]):
                    out += f"  |   {child[1]}: Node({child[2]})\n"
                out += "  |\n  -\n"
        return out

    def __iter__(self):
        """Return an iterator for each node below the current nav level."""
        iter_data = self.all_lower_nodes
        return NodeIterator(self, iter_data, delta=1)

    def __reversed__(self):
        """Return the iterator in reverse order."""
        iter_data = self.all_lower_nodes
        return NodeIterator(self, iter_data, delta=-1)

    def update_received(self, xmldoc):
        """Update nodes from event stream message."""
        address = value_from_xml(xmldoc, TAG_NODE)

        node = self.get_by_id(address)
        if not node:
            _LOGGER.debug(
                "Received a node update for node %s but could not find a record of this "
                "node. Please try restarting the module if the problem persists, this "
                "may be due to a new node being added to the ISY since last restart.",
                address,
            )
            return
        value = value_from_xml(xmldoc, ATTR_ACTION, "")
        value = int(value) if value != "" else ISY_VALUE_UNKNOWN
        prec = attr_from_xml(xmldoc, ATTR_ACTION, ATTR_PRECISION, DEFAULT_PRECISION)
        uom = attr_from_xml(
            xmldoc, ATTR_ACTION, ATTR_UNIT_OF_MEASURE, DEFAULT_UNIT_OF_MEASURE
        )
        formatted = value_from_xml(xmldoc, TAG_FORMATTED)

        # Process the action and value if provided in event data.
        node.update_state(
            NodeProperty(PROP_STATUS, value, prec, uom, formatted, address)
        )
        _LOGGER.debug("ISY Updated Node: %s", address)

    def control_message_received(self, xmldoc):
        """
        Pass Control events from an event stream message to nodes.

        Used for sending out to subscribers.
        """
        address = value_from_xml(xmldoc, TAG_NODE)
        cntrl = value_from_xml(xmldoc, ATTR_CONTROL)
        if not (address and cntrl):
            # If there is no node associated with the control message ignore it
            return

        node = self.get_by_id(address)
        if not node:
            _LOGGER.debug(
                "Received a node update for node %s but could not find a record of this "
                "node. Please try restarting the module if the problem persists, this "
                "may be due to a new node being added to the ISY since last restart.",
                address,
            )
            return

        # Process the action and value if provided in event data.
        node.update_last_update()
        value = value_from_xml(xmldoc, ATTR_ACTION, 0)
        value = int(value) if value != "" else ISY_VALUE_UNKNOWN
        prec = attr_from_xml(xmldoc, ATTR_ACTION, ATTR_PRECISION, DEFAULT_PRECISION)
        uom = attr_from_xml(
            xmldoc, ATTR_ACTION, ATTR_UNIT_OF_MEASURE, DEFAULT_UNIT_OF_MEASURE
        )
        formatted = value_from_xml(xmldoc, TAG_FORMATTED)

        if cntrl == PROP_RAMP_RATE:
            value = INSTEON_RAMP_RATES.get(value, value)
            uom = UOM_SECONDS
        node_property = NodeProperty(cntrl, value, prec, uom, formatted, address)
        if (
            cntrl == PROP_COMMS_ERROR
            and value == 0
            and PROP_COMMS_ERROR in node.aux_properties
        ):
            # Clear a previous comms error
            del node.aux_properties[PROP_COMMS_ERROR]
        if cntrl == PROP_BATTERY_LEVEL and node.is_battery_node:
            # Update the state if this is a battery node
            node.update_state(
                NodeProperty(PROP_STATUS, value, prec, uom, formatted, address)
            )
            _LOGGER.debug("ISY Updated Node: %s", address)
        elif cntrl not in EVENT_PROPS_IGNORED:
            node.update_property(node_property)
        node.control_events.notify(node_property)
        _LOGGER.debug("ISY Node Control Event: %s", node_property)

    def node_changed_received(self, xmldoc):
        """Handle Node Change/Update events from an event stream message."""
        action = value_from_xml(xmldoc, ATTR_ACTION)
        if not action or action not in NODE_CHANGED_ACTIONS:
            return
        (event_desc, e_i_keys) = NODE_CHANGED_ACTIONS[action]
        node = value_from_xml(xmldoc, TAG_NODE)
        detail = {}
        if e_i_keys and xmldoc.getElementsByTagName(TAG_EVENT_INFO):
            detail = {key: value_from_xml(xmldoc, key) for key in e_i_keys}

        if action == NC_NODE_ERROR:
            _LOGGER.error("ISY Could not communicate with device: %s", node)
        elif action == NC_NODE_ENABLED and node in self.addresses:
            node_obj: Node = self.get_by_id(node)
            # pylint: disable=attribute-defined-outside-init
            node_obj.enabled = detail[TAG_ENABLED] == XML_TRUE

        self.status_events.notify(event=NodeChangedEvent(node, action, detail))
        _LOGGER.debug(
            "ISY received a %s event for node %s %s",
            event_desc,
            node,
            detail if detail else "",
        )
        # FUTURE: Handle additional node change actions to force updates.

    def progress_report_received(self, xmldoc: minidom.Element) -> None:
        """Handle Progress Report '_7' events from an event stream message."""
        event_info = value_from_xml(xmldoc, TAG_EVENT_INFO)
        address, _, message = event_info.partition("]")
        address = address.strip("[ ")
        message = message.strip()
        action = DEV_WRITING
        detail = {"message": message}

        if address != "All" and message.startswith("Memory"):
            action = DEV_MEMORY
            regex = re.compile(MEMORY_REGEX)
            if event := regex.search(event_info):
                detail = {
                    "memory": event.group("dbAddr"),
                    "cmd1": event.group("cmd1"),
                    "cmd2": event.group("cmd2"),
                    "value": int(event.group("value"), 16),
                }
        self.status_events.notify(event=NodeChangedEvent(address, action, detail))
        _LOGGER.debug(
            "ISY received a progress report %s event for node %s %s",
            action,
            address,
            detail if detail else "",
        )

    def parse(self, xml):
        """
        Parse the xml data.

        |  xml: String of the xml data
        """
        try:
            xmldoc = minidom.parseString(xml)
        except XML_ERRORS as exc:
            _LOGGER.error("%s: Nodes", XML_PARSE_ERROR)
            raise ISYResponseParseError(XML_PARSE_ERROR) from exc

        # get nodes
        ntypes = [TAG_FOLDER, TAG_NODE, TAG_GROUP]
        node_servers = []
        for ntype in ntypes:
            features = xmldoc.getElementsByTagName(ntype)

            for feature in features:
                # Get Node Information
                address = value_from_xml(feature, TAG_ADDRESS)
                nname = value_from_xml(feature, TAG_NAME)

                _LOGGER.debug("Parsing %s: %s [%s]", ntype, nname, address)

                nparent = value_from_xml(feature, TAG_PARENT)
                pnode = value_from_xml(feature, TAG_PRIMARY_NODE)
                family = value_from_xml(feature, TAG_FAMILY)
                device_type = value_from_xml(feature, TAG_TYPE)
                node_def_id = attr_from_element(feature, ATTR_NODE_DEF_ID)
                flag = int(attr_from_element(feature, ATTR_FLAG, 0))
                enabled = value_from_xml(feature, TAG_ENABLED) == XML_TRUE

                # Assume Insteon, update as confirmed otherwise
                protocol = PROTO_INSTEON
                zwave_props = None
                node_server = None
                if family is not None:
                    if family in (FAMILY_ZWAVE, FAMILY_ZMATTER_ZWAVE):
                        protocol = PROTO_ZWAVE
                        zwave_prop_xml = feature.getElementsByTagName(TAG_DEVICE_TYPE)
                        if zwave_prop_xml:
                            zwave_props = ZWaveProperties.from_xml(zwave_prop_xml[0])
                        else:
                            ZWaveProperties()
                    elif family in (FAMILY_BRULTECH, FAMILY_RCS):
                        protocol = PROTO_ZIGBEE
                    elif family == FAMILY_NODESERVER:
                        # Node Server Slot is stored with family as text:
                        node_server = attr_from_xml(feature, TAG_FAMILY, ATTR_INSTANCE)
                        if node_server:
                            protocol = f"{PROTO_NODE_SERVER}_{node_server}"
                            node_servers.append(node_server)

                # Process the different node types
                if ntype == TAG_FOLDER and address not in self.addresses:
                    self.insert(address, nname, nparent, None, ntype)
                elif ntype == TAG_NODE:
                    if address in self.addresses:
                        self.get_by_id(address).update(xmldoc=feature)
                        continue
                    state, aux_props, state_set = parse_xml_properties(feature)
                    self.insert(
                        address,
                        nname,
                        nparent,
                        Node(
                            self,
                            address=address,
                            name=nname,
                            state=state,
                            aux_properties=aux_props,
                            zwave_props=zwave_props,
                            node_def_id=node_def_id,
                            pnode=pnode,
                            device_type=device_type,
                            enabled=enabled,
                            node_server=node_server,
                            protocol=protocol,
                            family_id=family,
                            state_set=state_set,
                            flag=flag,
                        ),
                        ntype,
                    )
                elif ntype == TAG_GROUP and address not in self.addresses:
                    # Ignore groups that contain 0x08 in the flag since
                    # that is a ISY scene that contains every device/
                    # scene so it will contain some scenes we have not
                    # seen yet so they are not defined and it includes
                    # the ISY MAC address in newer versions of
                    # ISY firmwares > 5.0.6+ ..
                    if flag & NODE_IS_ROOT:
                        _LOGGER.debug("Skipping root group flag=%s %s", flag, address)
                        continue
                    mems = feature.getElementsByTagName(TAG_LINK)
                    # Build list of members
                    members = [mem.firstChild.nodeValue for mem in mems]
                    # Build list of controllers
                    controllers = []
                    for mem in mems:
                        if (
                            int(attr_from_element(mem, TAG_TYPE, 0))
                            == NODE_IS_CONTROLLER
                        ):
                            controllers.append(mem.firstChild.nodeValue)
                    self.insert(
                        address,
                        nname,
                        nparent,
                        Group(
                            self,
                            address=address,
                            name=nname,
                            members=members,
                            controllers=controllers,
                            family_id=family,
                            pnode=pnode,
                            flag=flag,
                        ),
                        ntype,
                    )
            _LOGGER.debug("ISY Loaded %s", ntype)
        if self.isy.node_servers is None:
            self.isy.node_servers = NodeServers(self.isy, set(node_servers))

    async def update(self, wait_time=0, xml=None):
        """
        Update the status and properties of the nodes in the class.

        This calls the "/rest/status" endpoint.

        |  wait_time: [optional] Amount of seconds to wait before updating
        """
        if wait_time:
            await sleep(wait_time)

        if xml is None:
            xml = await self.isy.conn.get_status()

        if xml is None:
            _LOGGER.warning("ISY Failed to update nodes.")
            return

        try:
            xmldoc = minidom.parseString(xml)
        except XML_ERRORS:
            _LOGGER.error("%s: Nodes", XML_PARSE_ERROR)
            return False

        for feature in xmldoc.getElementsByTagName(TAG_NODE):
            address = feature.attributes[ATTR_ID].value

            if address in self.addresses:
                await self.get_by_id(address).update(xmldoc=feature)
                continue

        _LOGGER.info("ISY Updated Node Statuses.")

    async def update_nodes(self, wait_time=0):
        """
        Update the contents of the class.

        This calls the "/rest/nodes" endpoint.

        |  wait_time: [optional] Amount of seconds to wait before updating
        """
        if wait_time:
            await sleep(wait_time)
        xml = await self.isy.conn.get_nodes()
        if xml is None:
            _LOGGER.warning("ISY Failed to update nodes.")
            return
        self.parse(xml)

    def insert(self, address, nname, nparent, nobj, ntype):
        """
        Insert a new node into the lists.

        |  address: node id
        |  nname: node name
        |  nparent: node parent
        |  nobj: node object
        |  ntype: node type
        """
        self.addresses.append(address)
        self.nnames.append(nname)
        self.nparents.append(nparent)
        self.ntypes.append(ntype)
        self.nobjs.append(nobj)

    def __getitem__(self, val):
        """Navigate through the node tree. Can take names or IDs."""
        try:
            self.addresses.index(val)
            fun = self.get_by_id
        except ValueError:
            try:
                self.nnames.index(val)
                fun = self.get_by_name
            except ValueError:
                try:
                    val = int(val)
                    fun = self.get_by_index
                except ValueError:
                    fun = None

        if fun:
            output = None
            try:
                output = fun(val)
            except ValueError:
                pass

            if output:
                return output
        raise KeyError(f"Unrecognized Key: [{val}]")

    def __setitem__(self, item, value):
        """Set item value."""
        return None

    def get_by_name(self, val):
        """
        Get child object with the given name.

        |  val: String representing name to look for.
        """
        for i in range(len(self.addresses)):
            if (self.root is None or self.nparents[i] == self.root) and self.nnames[
                i
            ] == val:
                return self.get_by_index(i)
        return None

    def get_by_id(self, address):
        """
        Get object with the given ID.

        |  address: Integer representing node/group/folder id.
        """
        try:
            i = self.addresses.index(address)
        except ValueError:
            return None
        else:
            return self.get_by_index(i)

    def get_by_index(self, i):
        """
        Return the object at the given index in the list.

        |  i: Integer representing index of node/group/folder.
        """
        if self.ntypes[i] in [TAG_GROUP, TAG_NODE]:
            return self.nobjs[i]
        return Nodes(
            self.isy,
            self.addresses[i],
            self.addresses,
            self.nnames,
            self.nparents,
            self.nobjs,
            self.ntypes,
        )

    def get_folder(self, address):
        """Return the folder of a given node address."""
        parent = self.nparents[self.addresses.index(address)]
        if parent is None:
            # Node is in the root folder.
            return None
        parent_index = self.addresses.index(parent)
        if self.ntypes[parent_index] != TAG_FOLDER:
            return self.get_folder(parent)
        return self.nnames[parent_index]

    @property
    def children(self):
        """Return the children of the class."""
        return self.get_children()

    def get_children(self, ident=None):
        """Return the children of the class."""
        if ident is None:
            ident = self.root
        out = [
            (self.ntypes[i], self.nnames[i], self.addresses[i])
            for i in [
                index for index, parent in enumerate(self.nparents) if parent == ident
            ]
        ]
        return out

    @property
    def has_children(self):
        """Return if the root has children."""
        return self.root in self.nparents

    @property
    def name(self):
        """Return the name of the root."""
        if self.root is None:
            return ""
        ind = self.addresses.index(self.root)
        return self.nnames[ind]

    @property
    def all_lower_nodes(self):
        """Return all nodes below the current root."""
        output = []
        myname = self.name + "/"

        for dtype, name, ident in self.children:
            if dtype in [TAG_GROUP, TAG_NODE]:
                output.append((dtype, myname + name, ident))
                if dtype == TAG_NODE and ident in self.nparents:
                    output += [
                        (child[0], f"{myname}{name}/{child[1]}", child[2])
                        for child in self.get_children(ident)
                    ]
            if dtype == TAG_FOLDER:
                output += [
                    (dtype2, myname + name2, ident2)
                    for (dtype2, name2, ident2) in self[ident].all_lower_nodes
                ]
        return output


class NodeIterator:
    """Iterate through a list of nodes, returning node objects."""

    def __init__(self, nodes, iter_data, delta=1):
        """Initialize a NodeIterator class."""
        self._nodes = nodes
        self._iterdata = iter_data
        self._len = len(iter_data)
        self._delta = delta

        if delta > 0:
            self._ind = 0
        else:
            self._ind = self._len - 1

    def __next__(self):
        """Get the next element in the iteration."""
        if self._ind >= self._len or self._ind < 0:
            raise StopIteration
        _, path, ident = self._iterdata[self._ind]
        self._ind += self._delta
        return (path, self._nodes[ident])

    def __len__(self):
        """Return the number of elements."""
        return self._len


@dataclass
class NodeChangedEvent:
    """Class representation of a node change event."""

    address: str
    action: str
    event_info: dict



"""ISY Websocket Event Stream."""
import asyncio
import logging
import xml
from xml.dom import minidom

import aiohttp

from ..connection import get_new_client_session, get_sslcontext
from ..constants import (
    ACTION_KEY,
    ACTION_KEY_CHANGED,
    ATTR_ACTION,
    ATTR_CONTROL,
    ATTR_ID,
    ATTR_STREAM_ID,
    ATTR_VAR,
    ES_CONNECTED,
    ES_DISCONNECTED,
    ES_INITIALIZING,
    ES_LOST_STREAM_CONNECTION,
    ES_NOT_STARTED,
    ES_RECONNECTING,
    ES_STOP_UPDATES,
    PROP_STATUS,
    TAG_EVENT_INFO,
    TAG_NODE,
)
from ..helpers import attr_from_xml, now, value_from_xml
from ..logging import LOG_VERBOSE, enable_logging

_LOGGER = logging.getLogger(__name__)  # Allows targeting pyisy.events in handlers.

WS_HEADERS = {
    "Sec-WebSocket-Protocol": "ISYSUB",
    "Sec-WebSocket-Version": "13",
    "Origin": "com.universal-devices.websockets.isy",
}
WS_HEARTBEAT = 30
WS_HB_GRACE = 2
WS_TIMEOUT = 10.0
WS_MAX_RETRIES = 4
WS_RETRY_BACKOFF = [0.01, 1, 10, 30, 60]  # Seconds


class WebSocketClient:
    """Class for handling web socket communications with the ISY."""

    def __init__(
        self,
        isy,
        address,
        port,
        username,
        password,
        use_https=False,
        tls_ver=1.1,
        webroot="",
        websession=None,
    ):
        """Initialize a new Web Socket Client class."""
        if len(_LOGGER.handlers) == 0:
            enable_logging(add_null_handler=True)

        self.isy = isy
        self._address = address
        self._port = port
        self._username = username
        self._password = password
        self._auth = aiohttp.BasicAuth(self._username, self._password)
        self._webroot = webroot.rstrip("/")
        self._tls_ver = tls_ver
        self.use_https = use_https
        self._status = ES_NOT_STARTED
        self._lasthb = None
        self._hbwait = WS_HEARTBEAT
        self._sid = None
        self._program_key = None
        self.websocket_task = None
        self.guardian_task = None

        if websession is None:
            websession = get_new_client_session(use_https, tls_ver)

        self.req_session = websession
        self.sslcontext = get_sslcontext(use_https, tls_ver)
        self._loop = asyncio.get_running_loop()

        self._url = "wss://" if self.use_https else "ws://"
        self._url += f"{self._address}:{self._port}{self._webroot}/rest/subscribe"

    def start(self, retries=0):
        """Start the websocket connection."""
        if self.status != ES_CONNECTED:
            _LOGGER.debug("Starting websocket connection.")
            self.status = ES_INITIALIZING
            self.websocket_task = self._loop.create_task(self.websocket(retries))
            self.guardian_task = self._loop.create_task(self._websocket_guardian())

    def stop(self):
        """Close websocket connection."""
        self.status = ES_STOP_UPDATES
        if self.websocket_task is not None:
            _LOGGER.debug("Stopping websocket connection.")
            self.websocket_task.cancel()
        if self.guardian_task is not None:
            self.guardian_task.cancel()
            self._lasthb = None

    async def reconnect(self, delay=None, retries=0):
        """Reconnect to a disconnected websocket."""
        self.stop()
        self.status = ES_RECONNECTING
        if delay is None:
            delay = WS_RETRY_BACKOFF[retries]
        _LOGGER.info("PyISY attempting stream reconnect in %ss.", delay)
        await asyncio.sleep(delay)
        retries = (retries + 1) if retries < WS_MAX_RETRIES else WS_MAX_RETRIES
        self.start(retries)

    @property
    def status(self):
        """Return if the websocket is running or not."""
        return self._status

    @status.setter
    def status(self, value):
        """Set the current node state and notify listeners."""
        if self._status != value:
            self._status = value
            self.isy.connection_events.notify(self._status)
        return self._status

    @property
    def last_heartbeat(self):
        """Return the last received heartbeat time from the ISY."""
        return self._lasthb

    @property
    def heartbeat_time(self):
        """Return the time since the last ISY Heartbeat."""
        if self._lasthb is not None:
            return (now() - self._lasthb).seconds
        return 0.0

    async def _websocket_guardian(self):
        """Watch and reset websocket connection if no messages received."""
        while self.status != ES_STOP_UPDATES:
            await asyncio.sleep(self._hbwait)
            if (
                self.websocket_task.cancelled()
                or self.websocket_task.done()
                or self.heartbeat_time > self._hbwait + WS_HB_GRACE
            ):
                _LOGGER.debug("Websocket missed a heartbeat, resetting connection.")
                self.status = ES_LOST_STREAM_CONNECTION
                self._loop.create_task(self.reconnect())
                return

    async def _route_message(self, msg):
        """Route a received message from the event stream."""
        # check xml formatting
        try:
            xmldoc = minidom.parseString(msg)
        except xml.parsers.expat.ExpatError:
            _LOGGER.warning("ISY Received Malformed XML:\n%s", msg)
            return
        _LOGGER.log(LOG_VERBOSE, "ISY Update Received:\n%s", msg)

        # A wild stream id appears!
        if f"{ATTR_STREAM_ID}=" in msg and self._sid is None:
            self.update_received(xmldoc)

        # direct the event message
        cntrl = value_from_xml(xmldoc, ATTR_CONTROL)
        if not cntrl:
            return
        if cntrl == "_0":  # ISY HEARTBEAT
            self._lasthb = now()
            self._hbwait = int(value_from_xml(xmldoc, ATTR_ACTION))
            _LOGGER.debug("ISY HEARTBEAT: %s", self._lasthb.isoformat())
            self.isy.connection_events.notify(self._status)
        elif cntrl == PROP_STATUS:  # NODE UPDATE
            self.isy.nodes.update_received(xmldoc)
        elif cntrl[0] != "_":  # NODE CONTROL EVENT
            self.isy.nodes.control_message_received(xmldoc)
        elif cntrl == "_1":  # Trigger Update
            if f"<{ATTR_VAR}" in msg:  # VARIABLE (action=6 or 7)
                self.isy.variables.update_received(xmldoc)
            elif f"<{ATTR_ID}>" in msg:  # PROGRAM (action=0)
                self.isy.programs.update_received(xmldoc)
            elif f"<{TAG_NODE}>" in msg and "[" in msg:  # Node Server Update
                pass  # This is most likely a duplicate node update.
            elif f"<{ATTR_ACTION}>" in msg:
                action = value_from_xml(xmldoc, ATTR_ACTION)
                if action == ACTION_KEY:
                    self._program_key = value_from_xml(xmldoc, TAG_EVENT_INFO)
                    return
                if action == ACTION_KEY_CHANGED:
                    self._program_key = value_from_xml(xmldoc, TAG_NODE)
                # Need to reload programs
                await self.isy.programs.update()
        elif cntrl == "_3":  # Node Changed/Updated
            self.isy.nodes.node_changed_received(xmldoc)
        elif cntrl == "_5":  # System Status Changed
            self.isy.conn.system_status.change_received(xmldoc)
        elif cntrl == "_7":  # Progress report, device programming event
            self.isy.nodes.progress_report_received(xmldoc)

    def update_received(self, xmldoc):
        """Set the socket ID."""
        self._sid = attr_from_xml(xmldoc, "Event", ATTR_STREAM_ID)
        _LOGGER.debug("ISY Updated Events Stream ID: %s", self._sid)

    async def websocket(self, retries=0):
        """Start websocket connection."""
        try:
            async with self.req_session.ws_connect(
                self._url,
                auth=self._auth,
                heartbeat=WS_HEARTBEAT,
                headers=WS_HEADERS,
                timeout=WS_TIMEOUT,
                receive_timeout=self._hbwait + WS_HB_GRACE,
                ssl=self.sslcontext,
            ) as ws:
                self.status = ES_CONNECTED
                retries = 0
                _LOGGER.debug("Successfully connected to websocket.")

                async for msg in ws:
                    if msg.type == aiohttp.WSMsgType.TEXT:
                        await self._route_message(msg.data)
                    elif msg.type == aiohttp.WSMsgType.BINARY:
                        _LOGGER.warning("Unexpected binary message received.")
                    elif msg.type == aiohttp.WSMsgType.ERROR:
                        _LOGGER.error("Error during receive %s", ws.exception())
                        break

        except asyncio.CancelledError:
            self.status = ES_DISCONNECTED
            return
        except asyncio.TimeoutError:
            _LOGGER.debug("Websocket Timeout.")
        except aiohttp.ClientConnectorError as err:
            _LOGGER.error("Websocket Client Connector Error %s", err, exc_info=True)
        except (
            aiohttp.ClientOSError,
            aiohttp.client_exceptions.ServerDisconnectedError,
        ):
            _LOGGER.debug("Websocket Server Not Ready.")
        except aiohttp.client_exceptions.WSServerHandshakeError as err:
            _LOGGER.warning("Web socket server response error: %s", err.message)
        # pylint: disable=broad-except
        except Exception as err:
            _LOGGER.error("Unexpected websocket error %s", err, exc_info=True)
        else:
            if isinstance(ws.exception(), asyncio.TimeoutError):
                _LOGGER.debug("Websocket Timeout.")
            elif isinstance(ws.exception(), aiohttp.streams.EofStream):
                _LOGGER.warning(
                    "Websocket disconnected unexpectedly. Check network connection."
                )
            else:
                _LOGGER.warning(
                    "Websocket disconnected unexpectedly with code: %s", ws.close_code
                )
        if self.status != ES_STOP_UPDATES:
            self.status = ES_LOST_STREAM_CONNECTION
            self._loop.create_task(self.reconnect(retries=retries))



"""Helper functions for the PyISY Module."""
from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, is_dataclass
import datetime
import time

from .constants import (
    ATTR_FORMATTED,
    ATTR_ID,
    ATTR_PRECISION,
    ATTR_UNIT_OF_MEASURE,
    ATTR_VALUE,
    DEFAULT_PRECISION,
    DEFAULT_UNIT_OF_MEASURE,
    INSTEON_RAMP_RATES,
    ISY_EPOCH_OFFSET,
    ISY_PROP_NOT_SET,
    ISY_VALUE_UNKNOWN,
    PROP_BATTERY_LEVEL,
    PROP_RAMP_RATE,
    PROP_STATUS,
    TAG_CATEGORY,
    TAG_GENERIC,
    TAG_MFG,
    TAG_PROPERTY,
    UOM_SECONDS,
)
from .exceptions import XML_ERRORS
from .logging import _LOGGER


def parse_xml_properties(xmldoc):
    """
    Parse the xml properties string.

    Args:
        xmldoc: xml document to parse

    Returns:
        (state_val, state_uom, state_prec, aux_props)

    """
    aux_props = {}
    state_set = False
    state = NodeProperty(PROP_STATUS, uom=ISY_PROP_NOT_SET)

    props = xmldoc.getElementsByTagName(TAG_PROPERTY)
    if not props:
        return state, aux_props, state_set

    for prop in props:
        prop_id = attr_from_element(prop, ATTR_ID)
        uom = attr_from_element(prop, ATTR_UNIT_OF_MEASURE, DEFAULT_UNIT_OF_MEASURE)
        value = attr_from_element(prop, ATTR_VALUE, "").strip()
        prec = attr_from_element(prop, ATTR_PRECISION, DEFAULT_PRECISION)
        formatted = attr_from_element(prop, ATTR_FORMATTED, value)

        # ISY firmwares < 5 return a list of possible units.
        # ISYv5+ returns a UOM string which is checked against the SDK.
        # Only return a list if the UOM should be a list.
        if "/" in uom and uom != "n/a":
            uom = uom.split("/")

        value = int(value) if value.strip() != "" else ISY_VALUE_UNKNOWN

        result = NodeProperty(prop_id, value, prec, uom, formatted)

        if prop_id == PROP_STATUS:
            state = result
            state_set = True
        elif prop_id == PROP_BATTERY_LEVEL and not state_set:
            state = result
        else:
            if prop_id == PROP_RAMP_RATE:
                result.value = INSTEON_RAMP_RATES.get(value, value)
                result.uom = UOM_SECONDS
            aux_props[prop_id] = result

    return state, aux_props, state_set


def value_from_xml(xml, tag_name, default=None):
    """Extract a value from the XML element."""
    value = default
    try:
        value = xml.getElementsByTagName(tag_name)[0].firstChild.toxml()
    except XML_ERRORS:
        pass
    return value


def attr_from_xml(xml, tag_name, attr_name, default=None):
    """Extract an attribute value from the raw XML."""
    value = default
    try:
        root = xml.getElementsByTagName(tag_name)[0]
        value = attr_from_element(root, attr_name, default)
    except XML_ERRORS:
        pass
    return value


def attr_from_element(element, attr_name, default=None):
    """Extract an attribute value from an XML element."""
    value = default
    if attr_name in element.attributes.keys():
        value = element.attributes[attr_name].value
    return value


def value_from_nested_xml(base, chain, default=None):
    """Extract a value from multiple nested tags."""
    value = default
    result = None
    try:
        result = base.getElementsByTagName(chain[0])[0]
        if len(chain) > 1:
            result = result.getElementsByTagName(chain[1])[0]
        if len(chain) > 2:
            result = result.getElementsByTagName(chain[2])[0]
        if len(chain) > 3:
            result = result.getElementsByTagName(chain[3])[0]
        value = result.firstChild.toxml()
    except XML_ERRORS:
        pass
    return value


def ntp_to_system_time(timestamp):
    """Convert a ISY NTP time to system UTC time.

    Adapted from Python ntplib module.
    https://pypi.org/project/ntplib/

    Parameters:
    timestamp -- timestamp in NTP time

    Returns:
    corresponding system time

    Note: The ISY uses a EPOCH_OFFSET in addition to standard NTP.

    """
    _system_epoch = datetime.date(*time.gmtime(0)[0:3])
    _ntp_epoch = datetime.date(1900, 1, 1)
    ntp_delta = ((_system_epoch - _ntp_epoch).days * 24 * 3600) - ISY_EPOCH_OFFSET

    return datetime.datetime.fromtimestamp(timestamp - ntp_delta)


def now():
    """Get the current system time.

    Note: this module uses naive datetimes because the
    ISY is highly inconsistent with time conventions
    and does not present enough information to accurately
    manage DST without significant guessing and effort.
    """
    return datetime.datetime.now()


class EventEmitter:
    """Event Emitter class."""

    _subscribers: list[EventListener]

    def __init__(self):
        """Initialize a new Event Emitter class."""
        self._subscribers = []

    def subscribe(
        self, callback: Callable, event_filter: dict | str = None, key: str = None
    ):
        """Subscribe to the events."""
        listener = EventListener(
            emitter=self, callback=callback, event_filter=event_filter, key=key
        )
        self._subscribers.append(listener)
        return listener

    def unsubscribe(self, listener):
        """Unsubscribe from the events."""
        self._subscribers.remove(listener)

    def notify(self, event):
        """Notify a listener."""
        for subscriber in self._subscribers:
            # Guard against downstream errors interrupting the socket connection (#249)
            try:
                if e_filter := subscriber.event_filter:
                    if is_dataclass(event) and isinstance(e_filter, dict):
                        if not (e_filter.items() <= event.__dict__.items()):
                            continue
                    elif event != e_filter:
                        continue

                if subscriber.key:
                    subscriber.callback(event, subscriber.key)
                    continue
                subscriber.callback(event)
            except Exception:  # pylint: disable=broad-except
                _LOGGER.exception("Error during callback of %s", event)


@dataclass
class EventListener:
    """Event Listener class."""

    emitter: EventEmitter
    callback: Callable
    event_filter: dict | str
    key: str

    def unsubscribe(self):
        """Unsubscribe from the events."""
        self.emitter.unsubscribe(self)


@dataclass
class NodeProperty:
    """Class to hold result of a control event or node aux property."""

    control: str
    value: int | float = ISY_VALUE_UNKNOWN
    prec: str = DEFAULT_PRECISION
    uom: str = DEFAULT_UNIT_OF_MEASURE
    formatted: str = None
    address: str = None


@dataclass
class ZWaveProperties:
    """Class to hold Z-Wave Product Details from a Z-Wave Node."""

    category: str = "0"
    devtype_mfg: str = "0.0.0"
    devtype_gen: str = "0.0.0"
    basic_type: str = "0"
    generic_type: str = "0"
    specific_type: str = "0"
    mfr_id: str = "0"
    prod_type_id: str = "0"
    product_id: str = "0"
    raw: str = ""

    @classmethod
    def from_xml(cls, xml):
        """Return a Z-Wave Properties class from an xml DOM object."""
        category = value_from_xml(xml, TAG_CATEGORY)
        devtype_mfg = value_from_xml(xml, TAG_MFG)
        devtype_gen = value_from_xml(xml, TAG_GENERIC)
        raw = xml.toxml()
        basic_type = "0"
        generic_type = "0"
        specific_type = "0"
        mfr_id = "0"
        prod_type_id = "0"
        product_id = "0"
        if devtype_gen:
            (basic_type, generic_type, specific_type) = devtype_gen.split(".")
        if devtype_mfg:
            (mfr_id, prod_type_id, product_id) = devtype_mfg.split(".")

        return ZWaveProperties(
            category=category,
            devtype_mfg=devtype_mfg,
            devtype_gen=devtype_gen,
            basic_type=basic_type,
            generic_type=generic_type,
            specific_type=specific_type,
            mfr_id=mfr_id,
            prod_type_id=prod_type_id,
            product_id=product_id,
            raw=raw,
        )



"""ISY Network Resources Module."""
from asyncio import sleep
from xml.dom import minidom

from .constants import (
    ATTR_ID,
    PROTO_NETWORK,
    TAG_NAME,
    TAG_NET_RULE,
    URL_NETWORK,
    URL_RESOURCES,
)
from .exceptions import XML_ERRORS, XML_PARSE_ERROR
from .helpers import value_from_xml
from .logging import _LOGGER


class NetworkResources:
    """
    Network Resources class cobject.

    DESCRIPTION:
        This class handles the ISY networking module.

    USAGE:
        This object may be used in a similar way as a
        dictionary with the either networking command
        names or ids being used as keys and the ISY
        networking command class will be returned.

    EXAMPLE:
        # a = networking['test function']
        # a.run()

    ATTRIBUTES:
        isy: The ISY device class
        addresses: List of net command ids
        nnames: List of net command names
        nobjs: List of net command objects

    """

    def __init__(self, isy, xml=None):
        """
        Initialize the network resources class.

        isy: ISY class
        xml: String of xml data containing the configuration data
        """
        self.isy = isy

        self.addresses = []
        self.nnames = []
        self.nobjs = []

        if xml is not None:
            self.parse(xml)

    def parse(self, xml):
        """
        Parse the xml data.

        xml: String of the xml data
        """
        try:
            xmldoc = minidom.parseString(xml)
        except XML_ERRORS:
            _LOGGER.error("%s: NetworkResources, resources not loaded", XML_PARSE_ERROR)
            return

        features = xmldoc.getElementsByTagName(TAG_NET_RULE)
        for feature in features:
            address = int(value_from_xml(feature, ATTR_ID))
            if address not in self.addresses:
                nname = value_from_xml(feature, TAG_NAME)
                nobj = NetworkCommand(self, address, nname)
                self.addresses.append(address)
                self.nnames.append(nname)
                self.nobjs.append(nobj)

        _LOGGER.info("ISY Loaded Network Resources Commands")

    async def update(self, wait_time=0):
        """
        Update the contents of the networking class.

        wait_time: [optional] Amount of seconds to wait before updating
        """
        await sleep(wait_time)
        xml = await self.isy.conn.get_network()
        self.parse(xml)

    async def update_threaded(self, interval):
        """
        Continually update the class until it is told to stop.

        Should be run in a thread.
        """
        while self.isy.auto_update:
            await self.update(interval)

    def __getitem__(self, val):
        """Return the item from the collection."""
        try:
            val = int(val)
            return self.get_by_id(val)
        except (ValueError, KeyError):
            return self.get_by_name(val)

    def __setitem__(self, val, value):
        """Set the item value."""
        return None

    def get_by_id(self, val):
        """
        Return command object being given a command id.

        val: Integer representing command id
        """
        try:
            ind = self.addresses.index(val)
            return self.get_by_index(ind)
        except (ValueError, KeyError):
            return None

    def get_by_name(self, val):
        """
        Return command object being given a command name.

        val: String representing command name
        """
        try:
            ind = self.nnames.index(val)
            return self.get_by_index(ind)
        except (ValueError, KeyError):
            return None

    def get_by_index(self, val):
        """
        Return command object being given a command index.

        val: Integer representing command index in List
        """
        return self.nobjs[val]


class NetworkCommand:
    """
    Network Command Class.

    DESCRIPTION:
        This class handles individual networking commands.

    ATTRIBUTES:
        network_resources: The networkin resources class

    """

    def __init__(self, network_resources, address, name):
        """Initialize network command class.

        network_resources: NetworkResources class
        address: Integer of the command id
        """
        self._network_resources = network_resources
        self.isy = network_resources.isy
        self._id = address
        self._name = name

    @property
    def address(self):
        """Return the Resource ID for the Network Resource."""
        return str(self._id)

    @property
    def name(self):
        """Return the name of this entity."""
        return self._name

    @property
    def protocol(self):
        """Return the Protocol for this node."""
        return PROTO_NETWORK

    async def run(self):
        """Execute the networking command."""
        req_url = self.isy.conn.compile_url([URL_NETWORK, URL_RESOURCES, str(self._id)])

        if not await self.isy.conn.request(req_url, ok404=True):
            _LOGGER.warning("ISY could not run networking command: %s", str(self._id))
            return
        _LOGGER.debug("ISY ran networking command: %s", str(self._id))



"""ISY Variables."""
from asyncio import sleep
from xml.dom import minidom

from dateutil import parser

from ..constants import (
    ATTR_ID,
    ATTR_INIT,
    ATTR_PRECISION,
    ATTR_TS,
    ATTR_VAL,
    ATTR_VAR,
    TAG_NAME,
    TAG_TYPE,
    TAG_VARIABLE,
)
from ..exceptions import XML_ERRORS, XML_PARSE_ERROR, ISYResponseParseError
from ..helpers import attr_from_element, attr_from_xml, now, value_from_xml
from ..logging import _LOGGER
from .variable import Variable

EMPTY_VARIABLE_RESPONSES = [
    "/CONF/INTEGER.VAR not found",
    "/CONF/STATE.VAR not found",
    '<CList type="VAR_INT"></CList>',
]


class Variables:
    """
    This class handles the ISY variables.

    This class can be used as a     dictionary to navigate through the
    controller's structure to objects of type
    :class:`pyisy.variables.Variable` that represent objects on the
    controller.

    |  isy: The ISY object.
    |  root: The ID of the current level of navigation.
    |  vids: List of variable IDs from the controller.
    |  vnames: List of variable names form the controller.
    |  vobjs: List of variable objects.
    |  xml: XML string from the controller detailing the device's variables.

    :ivar children: List of the children below the current level of navigation.
    """

    def __init__(
        self,
        isy,
        root=None,
        vids=None,
        vnames=None,
        vobjs=None,
        def_xml=None,
        var_xml=None,
    ):
        """Initialize a Variables ISY Variable Manager class."""
        self.isy = isy
        self.root = root

        self.vids = {1: [], 2: []}
        self.vobjs = {1: {}, 2: {}}
        self.vnames = {1: {}, 2: {}}

        if vids is not None and vnames is not None and vobjs is not None:
            self.vids = vids
            self.vnames = vnames
            self.vobjs = vobjs
            return

        valid_definitions = False
        if def_xml is not None:
            valid_definitions = self.parse_definitions(def_xml)
        if valid_definitions and var_xml is not None:
            self.parse(var_xml)
        else:
            _LOGGER.warning("No valid variables defined")

    def __str__(self):
        """Return a string representation of the variable manager."""
        if self.root is None:
            return "Variable Collection"
        return f"Variable Collection (Type: {self.root})"

    def __repr__(self):
        """Return a string representing the children variables."""
        if self.root is None:
            return repr(self[1]) + repr(self[2])
        out = str(self) + "\n"
        for child in self.children:
            out += f"  {child[1]}: Variable({child[2]})\n"
        return out

    def parse_definitions(self, xmls):
        """Parse the XML Variable Definitions from the ISY."""
        valid_definitions = False
        for ind in range(2):
            # parse definitions
            if xmls[ind] is None or xmls[ind] in EMPTY_VARIABLE_RESPONSES:
                # No variables of this type defined.
                _LOGGER.info("No Type %s variables defined", ind + 1)
                continue
            try:
                xmldoc = minidom.parseString(xmls[ind])
            except XML_ERRORS:
                _LOGGER.error("%s: Type %s Variables", XML_PARSE_ERROR, ind + 1)
                continue
            else:
                features = xmldoc.getElementsByTagName(TAG_VARIABLE)
                for feature in features:
                    vid = int(attr_from_element(feature, ATTR_ID))
                    self.vnames[ind + 1][vid] = attr_from_element(feature, TAG_NAME)
                valid_definitions = True
        return valid_definitions

    def parse(self, xml):
        """Parse XML from the controller with details about the variables."""
        try:
            xmldoc = minidom.parseString(xml)
        except XML_ERRORS as exc:
            _LOGGER.error("%s: Variables", XML_PARSE_ERROR)
            raise ISYResponseParseError(XML_PARSE_ERROR) from exc

        features = xmldoc.getElementsByTagName(ATTR_VAR)
        for feature in features:
            vid = int(attr_from_element(feature, ATTR_ID))
            vtype = int(attr_from_element(feature, TAG_TYPE))
            init = value_from_xml(feature, ATTR_INIT)
            prec = int(value_from_xml(feature, ATTR_PRECISION, 0))
            val = value_from_xml(feature, ATTR_VAL)
            ts_raw = value_from_xml(feature, ATTR_TS)
            timestamp = parser.parse(ts_raw)
            vname = self.vnames[vtype].get(vid, "")

            vobj = self.vobjs[vtype].get(vid)
            if vobj is None:
                vobj = Variable(self, vid, vtype, vname, init, val, timestamp, prec)
                self.vids[vtype].append(vid)
                self.vobjs[vtype][vid] = vobj
            else:
                vobj.init = init
                vobj.status = val
                vobj.prec = prec
                vobj.last_edited = timestamp

        _LOGGER.info("ISY Loaded Variables")

    async def update(self, wait_time=0):
        """
        Update the variable objects with data from the controller.

        |  wait_time: Seconds to wait before updating.
        """
        await sleep(wait_time)
        xml = await self.isy.conn.get_variables()
        if xml is not None:
            self.parse(xml)
        else:
            _LOGGER.warning("ISY Failed to update variables.")

    def update_received(self, xmldoc):
        """Process an update received from the event stream."""
        xml = xmldoc.toxml()
        vtype = int(attr_from_xml(xmldoc, ATTR_VAR, TAG_TYPE))
        vid = int(attr_from_xml(xmldoc, ATTR_VAR, ATTR_ID))
        try:
            vobj = self.vobjs[vtype][vid]
        except KeyError:
            return  # this is a new variable that hasn't been loaded

        vobj.last_update = now()
        if f"<{ATTR_INIT}>" in xml:
            vobj.init = int(value_from_xml(xmldoc, ATTR_INIT))
        else:
            vobj.status = int(value_from_xml(xmldoc, ATTR_VAL))
            vobj.prec = int(value_from_xml(xmldoc, ATTR_PRECISION, 0))
            vobj.last_edited = parser.parse(value_from_xml(xmldoc, ATTR_TS))

        _LOGGER.debug("ISY Updated Variable: %s.%s", str(vtype), str(vid))

    def __getitem__(self, val):
        """
        Navigate through the variables by ID or name.

        |  val: Name or ID for navigation.
        """
        if self.root is None:
            if val in [1, 2]:
                return Variables(self.isy, val, self.vids, self.vnames, self.vobjs)
            raise KeyError(f"Unknown variable type: {val}")
        if isinstance(val, int):
            try:
                return self.vobjs[self.root][val]
            except (ValueError, KeyError) as err:
                raise KeyError(f"Unrecognized variable id: {val}") from err

        for vid, vname in self.vnames[self.root]:
            if vname == val:
                return self.vobjs[self.root][vid]
        raise KeyError(f"Unrecognized variable name: {val}")

    def __setitem__(self, val, value):
        """Handle the setitem function for the Class."""
        return None

    def get_by_name(self, val):
        """
        Get a variable with the given name.

        |  val: The name of the variable to look for.
        """
        vtype, _, vid = next(item for item in self.children if val in item)
        if not vid and vtype:
            raise KeyError(f"Unrecognized variable name: {val}")
        return self.vobjs[vtype].get(vid)

    @property
    def children(self):
        """Get the children of the class."""
        if self.root is None:
            types = [1, 2]
        else:
            types = [self.root]

        out = []
        for vtype in types:
            for ind in range(len(self.vids[vtype])):
                out.append(
                    (
                        vtype,
                        self.vnames[vtype].get(self.vids[vtype][ind], ""),
                        self.vids[vtype][ind],
                    )
                )
        return out



"""Representation of a node from an ISY."""
import asyncio
from math import isnan
from xml.dom import minidom

from ..constants import (
    BACKLIGHT_SUPPORT,
    CLIMATE_SETPOINT_MIN_GAP,
    CMD_CLIMATE_FAN_SETTING,
    CMD_CLIMATE_MODE,
    CMD_MANUAL_DIM_BEGIN,
    CMD_MANUAL_DIM_STOP,
    CMD_SECURE,
    INSTEON_SUBNODE_DIMMABLE,
    INSTEON_TYPE_DIMMABLE,
    INSTEON_TYPE_LOCK,
    INSTEON_TYPE_THERMOSTAT,
    METHOD_GET,
    METHOD_SET,
    PROP_ON_LEVEL,
    PROP_RAMP_RATE,
    PROP_SETPOINT_COOL,
    PROP_SETPOINT_HEAT,
    PROP_STATUS,
    PROP_ZWAVE_PREFIX,
    PROTO_INSTEON,
    PROTO_ZWAVE,
    TAG_CONFIG,
    TAG_GROUP,
    TAG_PARAMETER,
    TAG_SIZE,
    TAG_VALUE,
    UOM_CLIMATE_MODES,
    UOM_FAN_MODES,
    UOM_TO_STATES,
    URL_CONFIG,
    URL_NODE,
    URL_NODES,
    URL_QUERY,
    URL_ZWAVE,
    ZWAVE_CAT_DIMMABLE,
    ZWAVE_CAT_LOCK,
    ZWAVE_CAT_THERMOSTAT,
)
from ..exceptions import XML_ERRORS, XML_PARSE_ERROR, ISYResponseParseError
from ..helpers import (
    EventEmitter,
    NodeProperty,
    attr_from_xml,
    now,
    parse_xml_properties,
)
from ..logging import _LOGGER
from .nodebase import NodeBase


class Node(NodeBase):
    """
    This class handles ISY nodes.

    |  parent: The node manager object.
    |  address: The Node ID.
    |  value: The current Node value.
    |  name: The node name.
    |  spoken: The string of the Notes Spoken field.
    |  notes: Notes from the ISY
    |  uom: Unit of Measure returned by the ISY
    |  prec: Precision of the Node (10^-prec)
    |  aux_properties: Additional Properties for the node
    |  zwave_props: Z-Wave Properties from the devtype tag (used for Z-Wave Nodes.)
    |  node_def_id: Node Definition ID (used for ISY firmwares >=v5)
    |  pnode: Node ID of the primary node
    |  device_type: device type.
    |  node_server: the parent node server slot used
    |  protocol: the device protocol used (z-wave, zigbee, insteon, node server)

    :ivar status: A watched property that indicates the current status of the
                  node.
    :ivar has_children: Property indicating that there are no more children.
    """

    def __init__(
        self,
        nodes,
        address,
        name,
        state,
        aux_properties=None,
        zwave_props=None,
        node_def_id=None,
        pnode=None,
        device_type=None,
        enabled=None,
        node_server=None,
        protocol=None,
        family_id=None,
        state_set=True,
        flag=0,
    ):
        """Initialize a Node class."""
        self._enabled = enabled if enabled is not None else True
        self._formatted = state.formatted
        self._node_def_id = node_def_id
        self._node_server = node_server
        self._parent_node = pnode if pnode != address else None
        self._prec = state.prec
        self._protocol = protocol
        self._type = device_type
        self._uom = state.uom
        self._zwave_props = zwave_props
        self.control_events = EventEmitter()
        self._is_battery_node = not state_set
        super().__init__(
            nodes,
            address,
            name,
            state.value,
            family_id=family_id,
            aux_properties=aux_properties,
            pnode=pnode,
            flag=flag,
        )

    @property
    def dimmable(self):
        """
        Return the best guess if this is a dimmable node.

        DEPRECIATED: USE is_dimmable INSTEAD. Will be removed in future release.
        """
        _LOGGER.info("Node.dimmable is depreciated. Use Node.is_dimmable instead.")
        return self.is_dimmable

    @property
    def enabled(self):
        """Return if the device is enabled or not in the ISY."""
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        """Set if the device is enabled or not in the ISY."""
        if self._enabled != value:
            self._enabled = value

    @property
    def formatted(self):
        """Return the formatted value with units, if provided."""
        return self._formatted

    @property
    def is_battery_node(self):
        """
        Confirm if this is a battery node or a normal node.

        Battery nodes do not provide a 'ST' property, only 'BATLVL'.
        """
        return self._is_battery_node

    @property
    def is_backlight_supported(self) -> bool:
        """Confirm if this node supports setting backlight."""
        return (
            (self.protocol == PROTO_INSTEON)
            and self.node_def_id is not None
            and (self.node_def_id in BACKLIGHT_SUPPORT)
        )

    @property
    def is_dimmable(self) -> bool:
        """
        Return the best guess if this is a dimmable node.

        Check ISYv4 UOM, then Insteon and Z-Wave Types for dimmable types.
        """
        dimmable = (
            "%" in str(self._uom)
            or (
                self._protocol == PROTO_INSTEON
                and self.type
                and any({self.type.startswith(t) for t in INSTEON_TYPE_DIMMABLE})
                and self._id.endswith(INSTEON_SUBNODE_DIMMABLE)
            )
            or (
                self._protocol == PROTO_ZWAVE
                and self._zwave_props is not None
                and self._zwave_props.category in ZWAVE_CAT_DIMMABLE
            )
        )
        return dimmable

    @property
    def is_lock(self):
        """Determine if this device is a door lock type."""
        return (
            self.type and any({self.type.startswith(t) for t in INSTEON_TYPE_LOCK})
        ) or (
            self.protocol == PROTO_ZWAVE
            and self.zwave_props.category
            and self.zwave_props.category in ZWAVE_CAT_LOCK
        )

    @property
    def is_thermostat(self):
        """Determine if this device is a thermostat/climate control device."""
        return (
            self.type
            and any({self.type.startswith(t) for t in INSTEON_TYPE_THERMOSTAT})
        ) or (
            self._protocol == PROTO_ZWAVE
            and self.zwave_props.category
            and self.zwave_props.category in ZWAVE_CAT_THERMOSTAT
        )

    @property
    def node_def_id(self):
        """Return the node definition id (used for ISYv5)."""
        return self._node_def_id

    @property
    def node_server(self):
        """Return the node server parent slot (used for v5 Node Server devices)."""
        return self._node_server

    @property
    def parent_node(self):
        """
        Return the parent node object of this node.

        Typically this is for devices that are represented as multiple nodes in
        the ISY, such as door and leak sensors.
        Return None if there is no parent.

        """
        if self._parent_node:
            return self._nodes.get_by_id(self._parent_node)
        return None

    @property
    def prec(self):
        """Return the precision of the raw device value."""
        return self._prec

    @property
    def protocol(self):
        """Return the device standard used (Z-Wave, Zigbee, Insteon, Node Server)."""
        return self._protocol

    @property
    def type(self):
        """Return the device typecode (Used for Insteon)."""
        return self._type

    @property
    def uom(self):
        """Return the unit of measurement for the device."""
        return self._uom

    @property
    def zwave_props(self):
        """Return the Z-Wave Properties (used for Z-Wave devices)."""
        return self._zwave_props

    async def get_zwave_parameter(self, parameter):
        """Retrieve a Z-Wave Parameter from the ISY."""

        if not self.protocol == PROTO_ZWAVE:
            _LOGGER.warning("Cannot retrieve parameters of non-Z-Wave device")
            return

        if not isinstance(parameter, int):
            _LOGGER.error("Parameter must be an integer")
            return

        # /rest/zwave/node/<nodeAddress>/config/query/<parameterNumber>
        # returns something like:
        # <config paramNum="2" size="1" value="80"/>
        parameter_xml = await self.isy.conn.request(
            self.isy.conn.compile_url(
                [URL_ZWAVE, URL_NODE, self._id, URL_CONFIG, URL_QUERY, str(parameter)]
            )
        )

        if parameter_xml is None or parameter_xml == "":
            _LOGGER.warning("Error fetching parameter from ISY")
            return False

        try:
            parameter_dom = minidom.parseString(parameter_xml)
        except XML_ERRORS as exc:
            _LOGGER.error("%s: Node Parameter %s", XML_PARSE_ERROR, parameter_xml)
            raise ISYResponseParseError() from exc

        size = int(attr_from_xml(parameter_dom, TAG_CONFIG, TAG_SIZE))
        value = attr_from_xml(parameter_dom, TAG_CONFIG, TAG_VALUE)

        # Add/update the aux_properties to include the parameter.
        node_prop = NodeProperty(
            f"{PROP_ZWAVE_PREFIX}{parameter}",
            value,
            uom=f"{PROP_ZWAVE_PREFIX}{size}",
            address=self._id,
        )
        self.update_property(node_prop)

        return {TAG_PARAMETER: parameter, TAG_SIZE: size, TAG_VALUE: value}

    async def set_zwave_parameter(self, parameter, value, size):
        """Set a Z-Wave Parameter on an end device via the ISY."""

        if not self.protocol == PROTO_ZWAVE:
            _LOGGER.warning("Cannot set parameters of non-Z-Wave device")
            return False

        try:
            int(parameter)
        except ValueError:
            _LOGGER.error("Parameter must be an integer")
            return False

        if size not in [1, "1", 2, "2", 4, "4"]:
            _LOGGER.error("Size must either 1, 2, or 4 (bytes)")
            return False

        if str(value).startswith("0x"):
            try:
                int(value, base=16)
            except ValueError:
                _LOGGER.error("Value must be valid hex byte string or integer.")
                return False
        else:
            try:
                int(value)
            except ValueError:
                _LOGGER.error("Value must be valid hex byte string or integer.")
                return False

        # /rest/zwave/node/<nodeAddress>/config/set/<parameterNumber>/<value>/<size>
        req_url = self.isy.conn.compile_url(
            [
                URL_ZWAVE,
                URL_NODE,
                self._id,
                URL_CONFIG,
                METHOD_SET,
                str(parameter),
                str(value),
                str(size),
            ]
        )
        if not await self.isy.conn.request(req_url):
            _LOGGER.warning(
                "ISY could not set parameter %s on %s.",
                parameter,
                self._id,
            )
            return False
        _LOGGER.debug("ISY set parameter %s sent to %s.", parameter, self._id)

        # Add/update the aux_properties to include the parameter.
        node_prop = NodeProperty(
            f"{PROP_ZWAVE_PREFIX}{parameter}",
            value,
            uom=f"{PROP_ZWAVE_PREFIX}{size}",
            address=self._id,
        )
        self.update_property(node_prop)

        return True

    async def update(self, event=None, wait_time=0, xmldoc=None):
        """Update the value of the node from the controller."""
        if not self.isy.auto_update and not xmldoc:
            await asyncio.sleep(wait_time)
            req_url = self.isy.conn.compile_url(
                [URL_NODES, self._id, METHOD_GET, PROP_STATUS]
            )
            xml = await self.isy.conn.request(req_url)
            try:
                xmldoc = minidom.parseString(xml)
            except XML_ERRORS as exc:
                _LOGGER.error("%s: Nodes", XML_PARSE_ERROR)
                raise ISYResponseParseError(XML_PARSE_ERROR) from exc

        if xmldoc is None:
            _LOGGER.warning("ISY could not update node: %s", self._id)
            return

        self._last_update = now()
        state, aux_props, _ = parse_xml_properties(xmldoc)
        self._aux_properties.update(aux_props)
        self.update_state(state)
        _LOGGER.debug("ISY updated node: %s", self._id)

    def update_state(self, state):
        """Update the various state properties when received."""
        if not isinstance(state, NodeProperty):
            _LOGGER.error("Could not update state values. Invalid type provided.")
            return
        changed = False
        self._last_update = now()

        if state.prec != self._prec:
            self._prec = state.prec
            changed = True

        if state.uom not in (self._uom, ""):
            self._uom = state.uom
            changed = True

        if state.formatted != self._formatted:
            self._formatted = state.formatted
            changed = True

        if state.value != self.status:
            self.status = state.value
            # Let Status setter throw event
            return

        if changed:
            self._last_changed = now()
            self.status_events.notify(self.status_feedback)

    def get_command_value(self, uom, cmd):
        """Check against the list of UOM States if this is a valid command."""
        if cmd not in UOM_TO_STATES[uom].values():
            _LOGGER.warning(
                "Failed to call %s on %s, invalid command.", cmd, self.address
            )
            return None
        return list(UOM_TO_STATES[uom].keys())[
            list(UOM_TO_STATES[uom].values()).index(cmd)
        ]

    def get_groups(self, controller=True, responder=True):
        """
        Return the groups (scenes) of which this node is a member.

        If controller is True, then the scene it controls is added to the list
        If responder is True, then the scenes it is a responder of are added to
        the list.
        """
        groups = []
        for child in self._nodes.all_lower_nodes:
            if child[0] == TAG_GROUP:
                if responder:
                    if self._id in self._nodes[child[2]].members:
                        groups.append(child[2])
                elif controller:
                    if self._id in self._nodes[child[2]].controllers:
                        groups.append(child[2])
        return groups

    def get_property_uom(self, prop):
        """Get the Unit of Measurement an aux property."""
        if aux_prop := self._aux_properties.get(prop):
            return aux_prop.uom
        return None

    async def secure_lock(self):
        """Send a command to securely lock a lock device."""
        if not self.is_lock:
            _LOGGER.warning("Failed to lock %s, it is not a lock node.", self.address)
            return
        return await self.send_cmd(CMD_SECURE, "1")

    async def secure_unlock(self):
        """Send a command to securely lock a lock device."""
        if not self.is_lock:
            _LOGGER.warning("Failed to unlock %s, it is not a lock node.", self.address)
            return
        return await self.send_cmd(CMD_SECURE, "0")

    async def set_climate_mode(self, cmd):
        """Send a command to the device to set the climate mode."""
        if not self.is_thermostat:
            _LOGGER.warning(
                "Failed to set setpoint on %s, it is not a thermostat node.",
                self.address,
            )
        if cmd_value := self.get_command_value(UOM_CLIMATE_MODES, cmd):
            return await self.send_cmd(CMD_CLIMATE_MODE, cmd_value)
        return False

    async def set_climate_setpoint(self, val):
        """Send a command to the device to set the system setpoints."""
        if not self.is_thermostat:
            _LOGGER.warning(
                "Failed to set setpoint on %s, it is not a thermostat node.",
                self.address,
            )
            return
        adjustment = int(CLIMATE_SETPOINT_MIN_GAP / 2.0)

        commands = [
            self.set_climate_setpoint_heat(val - adjustment),
            self.set_climate_setpoint_cool(val + adjustment),
        ]
        result = await asyncio.gather(*commands, return_exceptions=True)
        return all(result)

    async def set_climate_setpoint_heat(self, val):
        """Send a command to the device to set the system heat setpoint."""
        return await self._set_climate_setpoint(val, "heat", PROP_SETPOINT_HEAT)

    async def set_climate_setpoint_cool(self, val):
        """Send a command to the device to set the system heat setpoint."""
        return await self._set_climate_setpoint(val, "cool", PROP_SETPOINT_COOL)

    async def _set_climate_setpoint(self, val, setpoint_name, setpoint_prop):
        """Send a command to the device to set the system heat setpoint."""
        if not self.is_thermostat:
            _LOGGER.warning(
                "Failed to set %s setpoint on %s, it is not a thermostat node.",
                setpoint_name,
                self.address,
            )
            return
        # ISY wants 2 times the temperature for Insteon in order to not lose precision
        if self._uom in ["101", "degrees"]:
            val = 2 * val
        return await self.send_cmd(
            setpoint_prop, str(val), self.get_property_uom(setpoint_prop)
        )

    async def set_fan_mode(self, cmd):
        """Send a command to the device to set the fan mode setting."""
        cmd_value = self.get_command_value(UOM_FAN_MODES, cmd)
        if cmd_value:
            return await self.send_cmd(CMD_CLIMATE_FAN_SETTING, cmd_value)
        return False

    async def set_on_level(self, val):
        """Set the ON Level for a device."""
        if not val or isnan(val) or int(val) not in range(256):
            _LOGGER.warning(
                "Invalid value for On Level for %s. Valid values are 0-255.", self._id
            )
            return False
        return await self.send_cmd(PROP_ON_LEVEL, str(val))

    async def set_ramp_rate(self, val):
        """Set the Ramp Rate for a device."""
        if not val or isnan(val) or int(val) not in range(32):
            _LOGGER.warning(
                "Invalid value for Ramp Rate for %s. "
                "Valid values are 0-31. See 'INSTEON_RAMP_RATES' in constants.py for values.",
                self._id,
            )
            return False
        return await self.send_cmd(PROP_RAMP_RATE, str(val))

    async def start_manual_dimming(self):
        """Begin manually dimming a device."""
        _LOGGER.warning(
            "'%s' is depreciated, use FADE__ commands instead", CMD_MANUAL_DIM_BEGIN
        )
        return await self.send_cmd(CMD_MANUAL_DIM_BEGIN)

    async def stop_manual_dimming(self):
        """Stop manually dimming  a device."""
        _LOGGER.warning(
            "'%s' is depreciated, use FADE__ commands instead", CMD_MANUAL_DIM_STOP
        )
        return await self.send_cmd(CMD_MANUAL_DIM_STOP)



"""ISY Configuration Lookup."""
from xml.dom import minidom

from .constants import (
    ATTR_DESC,
    ATTR_ID,
    TAG_DESC,
    TAG_FEATURE,
    TAG_FIRMWARE,
    TAG_INSTALLED,
    TAG_NAME,
    TAG_NODE_DEFS,
    TAG_PRODUCT,
    TAG_ROOT,
    TAG_VARIABLES,
    XML_TRUE,
)
from .exceptions import XML_ERRORS, XML_PARSE_ERROR, ISYResponseParseError
from .helpers import value_from_nested_xml, value_from_xml
from .logging import _LOGGER


class Configuration(dict):
    """
    ISY Configuration class.

    DESCRIPTION:
        This class handles the ISY configuration.

    USAGE:
        This object may be used in a similar way as a
        dictionary with the either module names or ids
        being used as keys and a boolean indicating
        whether the module is installed will be
        returned. With the exception of 'firmware' and 'uuid',
        which will return their respective values.

    PARAMETERS:
        Portal Integration - Check-it.ca
        Gas Meter
        SEP ESP
        Water Meter
        Z-Wave
        RCS Zigbee Device Support
        Irrigation/ETo Module
        Electricity Monitor
        AMI Electricity Meter
        URL
        A10/X10 for INSTEON
        Portal Integration - GreenNet.com
        Networking Module
        OpenADR
        Current Cost Meter
        Weather Information
        Broadband SEP Device
        Portal Integration - BestBuy.com
        Elk Security System
        Portal Integration - MobiLinc
        NorthWrite NOC Module

    EXAMPLE:
        # configuration['Networking Module']
        True
        # configuration['21040']
        True

    """

    def __init__(self, xml=None):
        """
        Initialize configuration class.

        xml: String of xml data containing the configuration data
        """
        super().__init__()

        if xml is not None:
            self.parse(xml)

    def parse(self, xml):
        """
        Parse the xml data.

        xml: String of the xml data
        """
        try:
            xmldoc = minidom.parseString(xml)
        except XML_ERRORS as exc:
            _LOGGER.error("%s: Configuration", XML_PARSE_ERROR)
            raise ISYResponseParseError(XML_PARSE_ERROR) from exc

        self["firmware"] = value_from_xml(xmldoc, TAG_FIRMWARE)
        self["uuid"] = value_from_nested_xml(xmldoc, [TAG_ROOT, ATTR_ID])
        self["name"] = value_from_nested_xml(xmldoc, [TAG_ROOT, TAG_NAME])
        self["model"] = value_from_nested_xml(xmldoc, [TAG_PRODUCT, TAG_DESC], "ISY")
        self["variables"] = bool(value_from_xml(xmldoc, TAG_VARIABLES) == XML_TRUE)
        self["nodedefs"] = bool(value_from_xml(xmldoc, TAG_NODE_DEFS) == XML_TRUE)

        features = xmldoc.getElementsByTagName(TAG_FEATURE)
        for feature in features:
            idnum = value_from_xml(feature, ATTR_ID)
            desc = value_from_xml(feature, ATTR_DESC)
            installed_raw = value_from_xml(feature, TAG_INSTALLED)
            installed = bool(installed_raw == XML_TRUE)
            self[idnum] = installed
            self[desc] = self[idnum]

        _LOGGER.info("ISY Loaded Configuration")



"""Implementation of module for command line.

The module can be tested by running the following command:
`python3 -m pyisy http://your-isy-url:80 username password`
Use `python3 -m pyisy -h` for full usage information.

This script can also be copied and used as a template for
using this module.
"""

import argparse
import asyncio
import logging
import time
from urllib.parse import urlparse

from . import ISY
from .connection import ISYConnectionError, ISYInvalidAuthError, get_new_client_session
from .constants import NODE_CHANGED_ACTIONS, SYSTEM_STATUS
from .logging import LOG_VERBOSE, enable_logging
from .nodes import NodeChangedEvent

_LOGGER = logging.getLogger(__name__)


async def main(url, username, password, tls_ver, events, node_servers):
    """Execute connection to ISY and load all system info."""
    _LOGGER.info("Starting PyISY...")
    t_0 = time.time()
    host = urlparse(url)
    if host.scheme == "http":
        https = False
        port = host.port or 80
    elif host.scheme == "https":
        https = True
        port = host.port or 443
    else:
        _LOGGER.error("host value in configuration is invalid.")
        return False

    # Use the helper function to get a new aiohttp.ClientSession.
    websession = get_new_client_session(https, tls_ver)

    # Connect to ISY controller.
    isy = ISY(
        host.hostname,
        port,
        username=username,
        password=password,
        use_https=https,
        tls_ver=tls_ver,
        webroot=host.path,
        websession=websession,
        use_websocket=True,
    )

    try:
        await isy.initialize(node_servers)
    except (ISYInvalidAuthError, ISYConnectionError):
        _LOGGER.error(
            "Failed to connect to the ISY, please adjust settings and try again."
        )
        await isy.shutdown()
        return
    except Exception as err:
        _LOGGER.error("Unknown error occurred: %s", err.args[0])
        await isy.shutdown()
        raise

    # Print a representation of all the Nodes
    _LOGGER.debug(repr(isy.nodes))
    _LOGGER.info("Total Loading time: %.2fs", time.time() - t_0)

    node_changed_subscriber = None
    system_status_subscriber = None

    def node_changed_handler(event: NodeChangedEvent) -> None:
        """Handle a node changed event sent from Nodes class."""
        (event_desc, _) = NODE_CHANGED_ACTIONS[event.action]
        _LOGGER.info(
            "Subscriber--Node %s Changed: %s %s",
            event.address,
            event_desc,
            event.event_info if event.event_info else "",
        )

    def system_status_handler(event: str) -> None:
        """Handle a system status changed event sent ISY class."""
        _LOGGER.info("System Status Changed: %s", SYSTEM_STATUS.get(event))

    try:
        if events:
            isy.websocket.start()
            node_changed_subscriber = isy.nodes.status_events.subscribe(
                node_changed_handler
            )
            system_status_subscriber = isy.system_status.status_events.subscribe(
                system_status_handler
            )
        while True:
            await asyncio.sleep(1)
    except asyncio.CancelledError:
        pass
    finally:
        if node_changed_subscriber:
            node_changed_subscriber.unsubscribe()
        if system_status_subscriber:
            system_status_subscriber.unsubscribe()
        await isy.shutdown()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog=__package__)
    parser.add_argument("url", type=str)
    parser.add_argument("username", type=str)
    parser.add_argument("password", type=str)
    parser.add_argument("-t", "--tls-ver", dest="tls_ver", type=float)
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("-q", "--no-events", dest="no_events", action="store_true")
    parser.add_argument(
        "-n", "--node-servers", dest="node_servers", action="store_true"
    )
    parser.set_defaults(use_https=False, tls_ver=1.1, verbose=False)
    args = parser.parse_args()

    enable_logging(LOG_VERBOSE if args.verbose else logging.DEBUG)

    _LOGGER.info(
        "ISY URL: %s, username: %s, TLS: %s",
        args.url,
        args.username,
        args.tls_ver,
    )

    try:
        asyncio.run(
            main(
                url=args.url,
                username=args.username,
                password=args.password,
                tls_ver=args.tls_ver,
                events=(not args.no_events),
                node_servers=args.node_servers,
            )
        )
    except KeyboardInterrupt:
        _LOGGER.warning("KeyboardInterrupt received. Disconnecting!")



"""ISY Node Server Information."""
import asyncio
from dataclasses import dataclass
import re
from typing import Dict, List
from xml.dom import getDOMImplementation, minidom

from .constants import (
    ATTR_ID,
    ATTR_UNIT_OF_MEASURE,
    TAG_ENABLED,
    TAG_NAME,
    TAG_ROOT,
    URL_PROFILE_NS,
)
from .exceptions import XML_ERRORS, XML_PARSE_ERROR, ISYResponseParseError
from .helpers import attr_from_element, value_from_xml
from .logging import _LOGGER

ATTR_DIR = "dir"
ATTR_EDITOR = "editor"
ATTR_NLS = "nls"
ATTR_SUBSET = "subset"
ATTR_PROFILE = "profile"

TAG_ACCEPTS = "accepts"
TAG_CMD = "cmd"
TAG_CONNECTION = "connection"
TAG_FILE = "file"
TAG_FILES = "files"
TAG_IP = "ip"
TAG_BASE_URL = "baseurl"
TAG_ISY_USER_NUM = "isyusernum"
TAG_NODE_DEF = "nodeDef"
TAG_NS_USER = "nsuser"
TAG_PORT = "port"
TAG_RANGE = "range"
TAG_SENDS = "sends"
TAG_SNI = "sni"
TAG_SSL = "ssl"
TAG_ST = "st"
TAG_TIMEOUT = "timeout"


class NodeServers:
    """
    ISY NodeServers class object.

    DESCRIPTION:
        This class handles the ISY Node Servers info.

    ATTRIBUTES:
        isy: The ISY device class

    """

    def __init__(self, isy, slots: list[str]):
        """
        Initialize the NodeServers class.

        isy: ISY class
        slots: List of slot numbers
        """
        self.isy = isy
        self._slots = slots
        self._connections = []
        self._profiles = {}
        self._node_server_node_definitions = []
        self._node_server_node_editors = []
        self._node_server_nls = []
        self.loaded = False

    async def load_node_servers(self):
        """Load information about node servers from the ISY."""

        await self.get_connection_info()
        await self.get_node_server_profiles()
        for slot in self._slots:
            await self.parse_node_server_defs(slot)
        self.loaded = True
        _LOGGER.info("ISY updated node servers")
        # _LOGGER.debug(self._node_server_node_definitions)
        # _LOGGER.debug(self._node_server_node_editors)

    async def get_connection_info(self):
        """Fetch the node server connections from the ISY."""
        result = await self.isy.conn.request(
            self.isy.conn.compile_url([URL_PROFILE_NS, "0", "connection"]),
            ok404=False,
        )
        if result is None:
            return

        try:
            connections_xml = minidom.parseString(result)
        except XML_ERRORS as exc:
            _LOGGER.error("%s while parsing Node Server connections", XML_PARSE_ERROR)
            raise ISYResponseParseError(XML_PARSE_ERROR) from exc

        connections = connections_xml.getElementsByTagName(TAG_CONNECTION)
        for connection in connections:
            self._connections.append(
                NodeServerConnection(
                    slot=attr_from_element(connection, ATTR_PROFILE),
                    enabled=attr_from_element(connection, TAG_ENABLED),
                    name=value_from_xml(connection, TAG_NAME),
                    ssl=value_from_xml(connection, TAG_SSL),
                    sni=value_from_xml(connection, TAG_SNI),
                    port=value_from_xml(connection, TAG_PORT),
                    timeout=value_from_xml(connection, TAG_TIMEOUT),
                    isy_user_num=value_from_xml(connection, TAG_ISY_USER_NUM),
                    ip=value_from_xml(connection, TAG_IP),
                    base_url=value_from_xml(connection, TAG_BASE_URL),
                    ns_user=value_from_xml(connection, TAG_NS_USER),
                )
            )
        _LOGGER.info("ISY updated node server connection info")

    async def get_node_server_profiles(self):
        """Retrieve the node server definition files from the ISY."""
        node_server_file_list = await self.isy.conn.request(
            self.isy.conn.compile_url([URL_PROFILE_NS, "0", "files"]), ok404=False
        )

        if node_server_file_list is None:
            return

        _LOGGER.debug("Parsing node server file list")

        try:
            file_list_xml = minidom.parseString(node_server_file_list)
        except XML_ERRORS as exc:
            _LOGGER.error("%s while parsing Node Server files", XML_PARSE_ERROR)
            raise ISYResponseParseError(XML_PARSE_ERROR) from exc

        file_list: List[str] = []

        profiles = file_list_xml.getElementsByTagName(ATTR_PROFILE)
        for profile in profiles:
            slot = attr_from_element(profile, ATTR_ID)
            directories = profile.getElementsByTagName(TAG_FILES)
            for directory in directories:
                dir_name = attr_from_element(directory, ATTR_DIR)
                files = directory.getElementsByTagName(TAG_FILE)
                for file in files:
                    file_name = attr_from_element(file, TAG_NAME)
                    file_list.append(f"{slot}/download/{dir_name}/{file_name}")

        file_tasks = [
            self.isy.conn.request(self.isy.conn.compile_url([URL_PROFILE_NS, file]))
            for file in file_list
        ]
        file_contents: List[str] = await asyncio.gather(*file_tasks)
        self._profiles: dict = dict(zip(file_list, file_contents))

        _LOGGER.info("ISY downloaded node server files")

    async def parse_node_server_defs(self, slot: str):
        """Retrieve and parse the node server definitions."""
        _LOGGER.info("Parsing node server slot %s", slot)
        node_server_profile = {
            key: value
            for (key, value) in self._profiles.items()
            if key.startswith(slot)
        }

        node_defs_impl = getDOMImplementation()
        editors_impl = getDOMImplementation()
        node_defs_xml = node_defs_impl.createDocument(None, TAG_ROOT, None)
        editors_xml = editors_impl.createDocument(None, TAG_ROOT, None)
        nls_lookup: dict = {}

        for file, contents in node_server_profile.items():
            contents_xml = ""
            file = file.lower()
            if file.endswith(".xml"):
                try:
                    contents_xml = minidom.parseString(contents).firstChild
                except XML_ERRORS:
                    _LOGGER.error(
                        "%s while parsing Node Server %s file %s",
                        XML_PARSE_ERROR,
                        slot,
                        file,
                    )
                    continue
            if "nodedef" in file:
                node_defs_xml.firstChild.appendChild(contents_xml)
            if "editors" in file:
                editors_xml.firstChild.appendChild(contents_xml)
            if "nls" in file and "en_us" in file:
                nls_list = [
                    line
                    for line in contents.split("\n")
                    if not line.startswith("#") and line != ""
                ]
                if nls_list:
                    nls_lookup = dict(re.split(r"\s?=\s?", line) for line in nls_list)
                    self._node_server_nls.append(
                        NodeServerNLS(
                            slot=slot,
                            nls=nls_lookup,
                        )
                    )

        # Process Node Def Files
        node_defs = node_defs_xml.getElementsByTagName(TAG_NODE_DEF)
        for node_def in node_defs:
            node_def_id = attr_from_element(node_def, ATTR_ID)
            nls_prefix = attr_from_element(node_def, ATTR_NLS)
            sts = node_def.getElementsByTagName(TAG_ST)
            statuses = {}
            for st in sts:
                status_id = attr_from_element(st, ATTR_ID)
                editor = attr_from_element(st, ATTR_EDITOR)
                statuses.update({status_id: editor})

            cmds_sends = node_def.getElementsByTagName(TAG_SENDS)[0]
            cmds_accepts = node_def.getElementsByTagName(TAG_ACCEPTS)[0]
            cmds_sends_cmd = cmds_sends.getElementsByTagName(TAG_CMD)
            cmds_accepts_cmd = cmds_accepts.getElementsByTagName(TAG_CMD)
            sends_commands = []
            accepts_commands = []

            for cmd in cmds_sends_cmd:
                sends_commands.append(attr_from_element(cmd, ATTR_ID))
            for cmd in cmds_accepts_cmd:
                accepts_commands.append(attr_from_element(cmd, ATTR_ID))

            status_names = {}
            name = node_def_id
            if nls_lookup:
                if (name_key := f"ND-{node_def_id}-NAME") in nls_lookup:
                    name = nls_lookup[name_key]
                for st in statuses:
                    if (key := f"ST-{nls_prefix}-{st}-NAME") in nls_lookup:
                        status_names.update({st: nls_lookup[key]})

            self._node_server_node_definitions.append(
                NodeServerNodeDefinition(
                    node_def_id=node_def_id,
                    name=name,
                    nls_prefix=nls_prefix,
                    slot=slot,
                    statuses=statuses,
                    status_names=status_names,
                    sends_commands=sends_commands,
                    accepts_commands=accepts_commands,
                )
            )
        # Process Editor Files
        editors = editors_xml.getElementsByTagName(ATTR_EDITOR)
        for editor in editors:
            editor_id = attr_from_element(editor, ATTR_ID)
            editor_range = editor.getElementsByTagName(TAG_RANGE)[0]
            uom = attr_from_element(editor_range, ATTR_UNIT_OF_MEASURE)
            subset = attr_from_element(editor_range, ATTR_SUBSET)
            nls = attr_from_element(editor_range, ATTR_NLS)

            values = None
            if nls_lookup and uom == "25":
                values = {
                    key.partition("-")[2]: value
                    for (key, value) in nls_lookup.items()
                    if key.startswith(nls)
                }

            self._node_server_node_editors.append(
                NodeServerNodeEditor(
                    editor_id=editor_id,
                    unit_of_measurement=uom,
                    subset=subset,
                    nls=nls,
                    slot=slot,
                    values=values,
                )
            )

        _LOGGER.debug("ISY parsed node server profiles")


@dataclass
class NodeServerNodeDefinition:
    """Node Server Node Definition parsed from the ISY/IoX."""

    node_def_id: str
    name: str
    nls_prefix: str
    slot: str
    statuses: Dict[str, str]
    status_names: Dict[str, str]
    sends_commands: List[str]
    accepts_commands: List[str]


@dataclass
class NodeServerNodeEditor:
    """Node Server Editor definition."""

    editor_id: str
    unit_of_measurement: str
    subset: str
    nls: str
    slot: str
    values: Dict[str, str]


@dataclass
class NodeServerNLS:
    """Node Server Natural Language Selection definition."""

    slot: str
    nls: Dict[str, str]


@dataclass
class NodeServerConnection:
    """Node Server Connection details."""

    slot: str
    enabled: str
    name: str
    ssl: str
    sni: str
    port: str
    timeout: str
    isy_user_num: str
    ip: str
    base_url: str
    ns_user: str

    def configuration_url(self) -> str:
        """Compile a configuration url from the connection data."""
        protocol: str = "https://" if self.ssl else "http://"
        return f"{protocol}{self.ip}:{self.port}"



"""Logging helper functions."""

import logging

_LOGGER = logging.getLogger(__package__)
LOG_LEVEL = logging.DEBUG
LOG_VERBOSE = 5
LOG_FORMAT = "%(asctime)s %(levelname)s [%(name)s] %(message)s"
LOG_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


def enable_logging(
    level=LOG_LEVEL,
    add_null_handler: bool = False,
    log_no_color: bool = False,
) -> None:
    """Set up the logging."""
    # Adapted from home-assistant/core/homeassistant/bootstrap.py
    if not log_no_color and not add_null_handler:
        try:
            # pylint: disable=import-outside-toplevel
            from colorlog import ColoredFormatter

            # basicConfig must be called after importing colorlog in order to
            # ensure that the handlers it sets up wraps the correct streams.
            logging.basicConfig(level=level)
            logging.addLevelName(LOG_VERBOSE, "VERBOSE")

            colorfmt = f"%(log_color)s{LOG_FORMAT}%(reset)s"
            logging.getLogger().handlers[0].setFormatter(
                ColoredFormatter(
                    colorfmt,
                    datefmt=LOG_DATE_FORMAT,
                    reset=True,
                    log_colors={
                        "VERBOSE": "blue",
                        "DEBUG": "cyan",
                        "INFO": "green",
                        "WARNING": "yellow",
                        "ERROR": "red",
                        "CRITICAL": "red",
                    },
                )
            )
        except ImportError:
            pass

    # If the above initialization failed for any reason, setup the default
    # formatting.  If the above succeeds, this will result in a no-op.
    logging.basicConfig(format=LOG_FORMAT, datefmt=LOG_DATE_FORMAT, level=level)

    if add_null_handler:
        _LOGGER.addHandler(logging.NullHandler())

    # Suppress overly verbose logs from libraries that aren't helpful
    logging.getLogger("requests").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("aiohttp.access").setLevel(logging.WARNING)



"""ISY Clock/Location Information."""
from asyncio import sleep
from xml.dom import minidom

from .constants import (
    EMPTY_TIME,
    TAG_DST,
    TAG_LATITUDE,
    TAG_LONGITUDE,
    TAG_MILIATRY_TIME,
    TAG_NTP,
    TAG_SUNRISE,
    TAG_SUNSET,
    TAG_TZ_OFFSET,
    XML_TRUE,
)
from .exceptions import XML_ERRORS, XML_PARSE_ERROR, ISYResponseParseError
from .helpers import ntp_to_system_time, value_from_xml
from .logging import _LOGGER


class Clock:
    """
    ISY Clock class cobject.

    DESCRIPTION:
        This class handles the ISY clock/location info.

    ATTRIBUTES:
        isy: The ISY device class
        last_called: the time of the last call to /rest/time
        tz_offset: The Time Zone Offset of the ISY
        dst: Daylight Savings Time Enabled or not
        latitude: ISY Device Latitude
        longitude: ISY Device Longitude
        sunrise: ISY Calculated Sunrise
        sunset: ISY Calculated Sunset
        military: If the clock is military time or not.

    """

    def __init__(self, isy, xml=None):
        """
        Initialize the network resources class.

        isy: ISY class
        xml: String of xml data containing the configuration data
        """
        self.isy = isy
        self._last_called = EMPTY_TIME
        self._tz_offset = 0
        self._dst = False
        self._latitude = 0.0
        self._longitude = 0.0
        self._sunrise = EMPTY_TIME
        self._sunset = EMPTY_TIME
        self._military = False

        if xml is not None:
            self.parse(xml)

    def __str__(self):
        """Return a string representing the clock Class."""
        return f"ISY Clock (Last Updated {self.last_called})"

    def __repr__(self):
        """Return a long string showing all the clock values."""
        props = [
            name for name, value in vars(Clock).items() if isinstance(value, property)
        ]
        return "ISY Clock: {!r}".format(
            {prop: str(getattr(self, prop)) for prop in props}
        )

    def parse(self, xml):
        """
        Parse the xml data.

        xml: String of the xml data
        """
        try:
            xmldoc = minidom.parseString(xml)
        except XML_ERRORS as exc:
            _LOGGER.error("%s: Clock", XML_PARSE_ERROR)
            raise ISYResponseParseError(XML_PARSE_ERROR) from exc

        tz_offset_sec = int(value_from_xml(xmldoc, TAG_TZ_OFFSET))
        self._tz_offset = tz_offset_sec / 3600
        self._dst = value_from_xml(xmldoc, TAG_DST) == XML_TRUE
        self._latitude = float(value_from_xml(xmldoc, TAG_LATITUDE))
        self._longitude = float(value_from_xml(xmldoc, TAG_LONGITUDE))
        self._military = value_from_xml(xmldoc, TAG_MILIATRY_TIME) == XML_TRUE
        self._last_called = ntp_to_system_time(int(value_from_xml(xmldoc, TAG_NTP)))
        self._sunrise = ntp_to_system_time(int(value_from_xml(xmldoc, TAG_SUNRISE)))
        self._sunset = ntp_to_system_time(int(value_from_xml(xmldoc, TAG_SUNSET)))

        _LOGGER.info("ISY Loaded Clock Information")

    async def update(self, wait_time=0):
        """
        Update the contents of the networking class.

        wait_time: [optional] Amount of seconds to wait before updating
        """
        await sleep(wait_time)
        xml = await self.isy.conn.get_time()
        self.parse(xml)

    async def update_thread(self, interval):
        """
        Continually update the class until it is told to stop.

        Should be run as a task in the event loop.
        """
        while self.isy.auto_update:
            await self.update(interval)

    @property
    def last_called(self):
        """Get the time of the last call to /rest/time in UTC."""
        return self._last_called

    @property
    def tz_offset(self):
        """Provide the Time Zone Offset from the isy in Hours."""
        return self._tz_offset

    @property
    def dst(self):
        """Confirm if DST is enabled or not on the ISY."""
        return self._dst

    @property
    def latitude(self):
        """Provide the latitude information from the isy."""
        return self._latitude

    @property
    def longitude(self):
        """Provide the longitude information from the isy."""
        return self._longitude

    @property
    def sunrise(self):
        """Provide the sunrise information from the isy (UTC)."""
        return self._sunrise

    @property
    def sunset(self):
        """Provide the sunset information from the isy (UTC)."""
        return self._sunset

    @property
    def military(self):
        """Confirm if military time is in use or not on the isy."""
        return self._military
